#ifndef __TCOLOUR_HH
#define __TCOLOUR_HH
//
// Project : lib3d
// File    : TColour.hh
// Purpose : Declaration of a class for colour-management
// Author  : Ronald Kriemann
//
// arch-tag: 2d2fc7bf-61ae-4b26-9c7c-cb8cc8ac5154
//

#include <iostream>

#include "defines.hh"
#include "TString.hh"
#include "TColourLookup.hh"

namespace RGL
{

// adjust colour-value c to be in [0,1]
#define COL_ADJUST( c ) ((c) > 1.0 ? 1.0 : ((c) < 0.0 ? 0.0 : (c)))

//
// class for colour-management
// (default = black)
//
class TColour
{
protected:
    // colour values ( 0 <= r,g,b <= 1 )
    real   _red, _green, _blue;

public:
    ///////////////////////////////////////////////////
    //
    // constructors and destructor
    //

    TColour () : _red(0), _green(0), _blue(0) {}
    TColour ( const TColour & c )             { *this = c; }
    TColour ( real r, real g, real b )        { set_rgb( r, g, b ); }
    TColour ( const TString & name )          { set_named( name ); }

    ///////////////////////////////////////////////////
    //
    // access colour values
    //

    // return colour components
    real red   () const { return _red; }
    real green () const { return _green; }
    real blue  () const { return _blue; }

    // set to given (scaled) colour
    TColour & set ( const TColour & c ) { *this = c; return *this; }
    TColour & set ( real f, const TColour & c, bool check = true )
    {
        set_rgb( f * c._red, f * c._green, f * c._blue, check );
        return *this;
    }
    
    // set by RGB values
    TColour & set_rgb ( real r, real g, real b, bool check = true )
    {
        if ( check )
        {
            _red   = COL_ADJUST(r);
            _green = COL_ADJUST(g);
            _blue  = COL_ADJUST(b);
        }// if
        else
        {
            _red   = r;
            _green = g;
            _blue  = b;
        }// else

        return *this;
    }
    
    // get/set colour in HSV-space ( h \in [0,360] \cup {-1}, s,v \in [0,1] )
    void      hsv     ( real & h, real & s, real & v ) const;
    TColour & set_hsv ( real   h, real   s, real   v );

    // set colour as named-value
    TColour & set_named ( const TString & name );
    
    // add a (scaled) colour
    TColour & add ( const TColour & col, bool check = true )
    {
        set_rgb( _red + col._red, _green + col._green, _blue + col._blue, check );
        return *this;
    }

    TColour & add ( real f, const TColour & col, bool check = true )
    {
        set_rgb( _red + f * col._red, _green + f * col._green, _blue + f * col._blue, check );
        return *this;
    }

    // scale colour by factor
    TColour & scale ( real f, bool check = true )
    {
        set_rgb( _red * f, _green * f, _blue * f, check );
        return *this;
    }

    // set colour from 32 bit integer (from lsb to msb: red,green,blue,alpha)
    TColour & set_rgb32 ( uint rgb );

    ///////////////////////////////////////////////////
    //
    // misc. operators
    //

    // copy
    TColour & operator = ( const TColour & c )
    {
        _red = c._red; _green = c._green; _blue = c._blue;
        return *this;
    }
    
    // compare
    bool operator == ( const TColour & c ) const
    { return ((_red == c._red) && (_green == c._green) && (_blue == c._blue)); }
    
    bool operator != ( const TColour & c ) const { return ! (*this == c); }

    // stream output
    friend std::ostream & operator << ( std::ostream & os, const TColour & c );
};

} // namespace

#endif  // __TCOLOUR_HH
