#ifndef __TIMAGE_HH
#define __TIMAGE_HH
//
// Project : lib3D
// File    : TImage.hh
// Purpose : class for hardware-independent images
// Author  : Ronald Kriemann
//
// arch-tag: 9d96b31b-6350-4261-b91b-92c41f133ac0
//

#include "defines.hh"

#include "TArray.hh"
#include "TColour.hh"

namespace RGL
{

//
// hardware-independet representation of images
// - pixels are represented by TColour type
// - able to load images from files with given loader
//
class TImage
{
protected:
    // dimension of image
    uint              _width, _height;

    // the image data (for different depths)
    TArray< TColour > _image;

public:
    //////////////////////////////////////////////////////
    //
    // constructor and destructor
    //

    TImage () : _width(0), _height(0) {}
    TImage ( uint w, uint h ) : _width(w), _height(h) { _image.set_size( _width * _height ); }

    virtual ~TImage () {}
    
    //////////////////////////////////////////////////////
    //
    // access local variables
    //

    uint  width  () const { return _width; }
    uint  height () const { return _height; }

    // set new width and height without scaling
    virtual void set_geometry ( uint w, uint h );
    
    // access colours at pos (i,j)
    const TColour & get_colour ( uint i, uint j ) const { return (*this)(i,j); }
    void            set_colour ( uint i, uint j, const TColour & c ) { (*this)(i,j) = c; }

    const TColour & operator () ( uint i, uint j ) const
    { assert((i < _width) && (j < _height)); return _image[ (j * _width) + i ]; }

    TColour &       operator () ( uint i, uint j )
    { assert((i < _width) && (j < _height)); return _image[ (j * _width) + i ]; }

    // access pixel array
    const TArray< TColour > & pix_array () const { return _image; }
    TArray< TColour >       & pix_array ()       { return _image; }

    //////////////////////////////////////////////////
    //
    // some primitives
    //

    // draw a point
    virtual void draw_point ( uint x, uint y, const TColour & c ) { set_colour( x, y, c ); }
    
    // fill image with colour
    virtual void fill ( const TColour & c );

    // copy given image at given position
    virtual void bitblt ( uint x, uint y, const TImage & img );
    
    //////////////////////////////////////////////////////
    //
    // image-manipulation
    //

    // change colour values to grayscale
    virtual void to_grayscale ();

    // make image lighter / darker
    virtual void darker  ();
    virtual void lighter ();
    
    //////////////////////////////////////////////////////
    //
    // misc.
    //

    // copy methods
    TImage & copy       ( const TImage & img );
    TImage & operator = ( const TImage & img ) { return copy( img ); }

    // return size in bytes the image uses
    ulong byte_size () const { return sizeof(TColour) * _width * _height; }
};

} // namespace

#endif  // __TIMAGE_HH
