#ifndef __TMATRIX_HH
#define __TMATRIX_HH
//
// Project : lib3D
// File    : TMatrix.hh
// Purpose : Declaration of a class for a 4x4 matrix (used in 3d-transformation)
// Author  : Ronald Kriemann
//
// arch-tag: ceb05b31-8e7e-46cc-879d-bbcee659ec07
//

#include <iostream>

#include "defines.hh"

#include "TVector.hh"

namespace RGL
{

// define for usual matrix-access
#define M(i,j)  _data[ (4*i) + j ]

//
// fully populated :-) 4x4 matrix
// (last row is implicitly set to [ 0 0 0 1 ])
//
class TMatrix
{
protected:
    // matrix coefficients
    real  _data[12];

public:
    ///////////////////////////////////////////////
    //
    // constructors and destructors
    //

    TMatrix ()                    { set_identity(); }
    TMatrix ( real f )            { for ( uint i = 0; i < 12; i++ ) _data[i] = f; }
    TMatrix ( const TMatrix & m ) { *this = m; }

    virtual ~TMatrix () {}

    ///////////////////////////////////////////////
    //
    // access coefficients
    //

    real & operator () ( uint i, uint j )       { assert((i<3) && (j<4)); return M(i,j); }
    real   operator () ( uint i, uint j ) const { assert((i<3) && (j<4)); return M(i,j); }

    real   entry       ( uint i, uint j ) const { assert((i<3) && (j<4)); return M(i,j); }
    
    // set to identity
    TMatrix & set_identity ()
    {
        _data[0] = 1.0; _data[1] = 0.0; _data[2]  = 0.0; _data[3]  = 0.0; 
        _data[4] = 0.0; _data[5] = 1.0; _data[6]  = 0.0; _data[7]  = 0.0; 
        _data[8] = 0.0; _data[9] = 0.0; _data[10] = 1.0; _data[11] = 0.0;
        return *this;
    }

    // set to null
    TMatrix & set_null ()
    {
        for ( uint i = 0; i < 12; i++ ) _data[i] = 0.0;
        return *this;
    }

    ///////////////////////////////////////////////
    //
    // usual matrix methods
    //

    // multiply given matrix from right/left
    TMatrix & mul_right ( const TMatrix & m );
    TMatrix & mul_left  ( const TMatrix & m );
    
    // x = *this y (full or upper 3x3 matrix)
    void mul_vec  ( const TVector & y, TVector & x ) const
    {
        // must use temp. vector in case of x = M x
        const TVector  t( y );
     
        x[0] = ( _data[0]  * t._data[0] + _data[1]  * t._data[1] +
                 _data[2]  * t._data[2] + _data[3] );
        x[1] = ( _data[4]  * t._data[0] + _data[5]  * t._data[1] +
                 _data[6]  * t._data[2] + _data[7] );
        x[2] = ( _data[8]  * t._data[0] + _data[9]  * t._data[1] +
                 _data[10] * t._data[2] + _data[11] );
    }

    void mul_vec3 ( const TVector & y, TVector & x, bool trans = false ) const
    {
        // must use temp. vector in case of x = M x
        const TVector  t( y );
     
        if ( trans )
        {
            x[0] = ( _data[0] * t._data[0] + _data[4] * t._data[1] + _data[8]  * t._data[2] );
            x[1] = ( _data[1] * t._data[0] + _data[5] * t._data[1] + _data[9]  * t._data[2] );
            x[2] = ( _data[2] * t._data[0] + _data[6] * t._data[1] + _data[10] * t._data[2] );
        }// if
        else
        {
            x[0] = ( _data[0] * t._data[0] + _data[1] * t._data[1] + _data[2]  * t._data[2] );
            x[1] = ( _data[4] * t._data[0] + _data[5] * t._data[1] + _data[6]  * t._data[2] );
            x[2] = ( _data[8] * t._data[0] + _data[9] * t._data[1] + _data[10] * t._data[2] );
        }// else
    }

    // x = *this x
    TVector & mul_vec  ( TVector & y ) const
    {
        const TVector  t( y );
        
        y[0] = ( _data[0]  * t._data[0] + _data[1]  * t._data[1] +
                 _data[2]  * t._data[2] + _data[3] );
        y[1] = ( _data[4]  * t._data[0] + _data[5]  * t._data[1] +
                 _data[6]  * t._data[2] + _data[7] );
        y[2] = ( _data[8]  * t._data[0] + _data[9]  * t._data[1] +
                 _data[10] * t._data[2] + _data[11] );
        
        return y;
    }

    // invert matrix: this = this^-1
    TMatrix & invert ();

    // invert matrix: C = this^-1
    void invert  ( TMatrix & C ) const;

    // invert upper 3x3 part of matrix to C
    void invert3 ( TMatrix & C ) const;

    ///////////////////////////////////////////////
    //
    // set matrix to various transformations
    //

    // build scaling by factor <sx, sy, sz>
    TMatrix & set_scale     ( const TVector & s );

    // build translation by vector <tx, ty, tz>
    TMatrix & set_translate ( const TVector & t );

    // build rotation around v and angle alpha
    TMatrix & set_rotate    ( const TVector & v, real alpha );
    
    /////////////////////////////////////////////
    //
    // misc methods and operators
    //

    TMatrix & operator = ( const TMatrix & m )
    {
        for ( uint i = 0; i < 12; i++ ) _data[i] = m._data[i];
        return *this;
    }

    friend std::ostream & operator << ( std::ostream & os, const TMatrix & m );
};

// reset matrix-defines
#undef M

} // namespace

#endif  // __TMATRIX_HH
