#ifndef __TRAY_HH
#define __TRAY_HH
//
// Project : lib3D
// File    : TRay.hh
// Purpose : class representing a ray
// Author  : Ronald Kriemann
//
// arch-tag: 618ee74e-3fa4-4833-bd26-4f45e84becc0
//

#include <iostream>

#include "TVector.hh"
#include "TMatrix.hh"

namespace RGL
{

// maximal number of history entries
#define MAX_HISTORY  100

// forward decl.
class TGraphObject;

//
// ray is defined by x + t * d
// where d = y - x
//
class TRay
{
private:
    // ID counter
    static ulong         _id_counter;
    
protected:
    // the origin and direction of the ray
    TVector              _origin, _dir;

    // ID if the ray
    ulong                _id;

    // true if initial ray
    bool                 _initial;

    // history of entered objects
    const TGraphObject * _history[ MAX_HISTORY ];
    int                  _hist_idx;
    
public:
    ////////////////////////////////////////////////
    //
    // constructor and destructor
    //

    TRay ( bool initial = false )
            : _id(_id_counter++), _initial(initial),
              _hist_idx(-1)
    {}
    
    TRay ( const TVector & origin,
           const TVector & dir,
           bool            initial = false )
            : _origin( origin ), _dir( dir ), _id(_id_counter++),
              _initial(initial), _hist_idx(-1)
    {}
    
    TRay ( const TRay & ray ) { *this = ray; }

    ////////////////////////////////////////////////
    //
    // access coordinates
    //

    TVector & origin    () { return _origin; }
    TVector & direction () { return _dir; }
    
    const TVector & origin    () const { return _origin; }
    const TVector & direction () const { return _dir; }
    ulong           id        () const { return _id; }

    TRay & set_origin    ( const TVector & v ) { _origin = v; return *this; }
    TRay & set_look_at   ( const TVector & v ) { _dir.set( 1.0, v, -1.0, _origin ); return *this; }
    TRay & set_direction ( const TVector & v ) { _dir = v; return *this; }
    
    TRay & set ( const TVector & o, const TVector & d )
    { _origin = o; _dir = d; return *this; }

    bool is_initial  () const { return _initial; }
    void set_initial ()       { _initial = true; }
    
    ////////////////////////////////////////////////
    //
    // history management
    //

    // return true if ray is not inside objects
    bool outside_objects () const { return _hist_idx == -1; }

    // return current object ray is inside
    const TGraphObject * current_object () const
    {
        if ( _hist_idx == -1 ) return NULL;
        else                   return _history[ _hist_idx ];
    }
    
    // enter object through history
    void enter ( const TGraphObject * obj )
    {
        if ( (++_hist_idx) >= MAX_HISTORY )
            std::cout << "(TRay) enter : too many nested refracting objects"
                      << std::endl;
        _history[_hist_idx] = obj;
    }

    // exit object number 'nr'
    void exit ( int nr )
    {
        for ( int i = nr; i < _hist_idx; i++ )
            _history[i] = _history[i+1];

        if ( (--_hist_idx) < -1 )
            std::cout << "(TRay) exit : too many exits from refracting objects"
                      << std::endl;
    }

    // check if ray has entered given object
    bool has_entered ( const TGraphObject * obj, int & nr ) const
    {
        if ( _hist_idx > -1 )
        {
            for ( int i = 0; i < _hist_idx; i++ )
            {
                if ( _history[i] == obj )
                {
                    nr = i;
                    return true;
                }// if
            }// for
        }// if

        return false;
    }
    
    ////////////////////////////////////////////////
    //
    // misc.
    //

    // set new ID to ray
    void new_id () { _id = _id_counter++; }
    
    // copy
    TRay & operator = ( const TRay & ray )
    {
        _origin  = ray._origin;
        _dir     = ray._dir;
        _id      = ray._id;
        _initial = ray._initial;

        _hist_idx = ray._hist_idx;
        
        for ( int i = 0; i <= _hist_idx; i++ )
            _history[i] = ray._history[i];

        return *this;
    }

    // copy history from given ray
    void copy_history ( const TRay & ray )
    {
        _hist_idx = ray._hist_idx;
        for ( int i = 0; i <= _hist_idx; i++ )
            _history[i] = ray._history[i];
    }
    
    // transform ray by given matrix
    void transform ( const TMatrix & M, TRay & ray ) const
    {
        M.mul_vec( _origin, ray._origin );
        // only upper 3x3 because (destination - origin).w() == 0
        M.mul_vec3( _dir, ray._dir );
    }
    
    // stream output
    friend std::ostream & operator << ( std::ostream & os, const TRay & ray )
    {
        os << ray._id << " : " << ray.origin() << " -> " << ray.direction()
           << std::endl;
        return os;
    }
};

} // namespace

#endif  // __TRAY_HH
