#ifndef __TTEXTURE_HH
#define __TTEXTURE_HH
//
// Project : lib3D
// File    : TTexture.hh
// Purpose : (base)class for describing objects
// Author  : Ronald Kriemann
//
// arch-tag: 69430739-3adb-4f24-8b2e-efc05ecb2abc
//

#include "TColour.hh"
#include "TMultiRef.hh"
#include "TVector.hh"

namespace RGL
{

//
// holds all data to display the colour or
// surface of objects
// - a printing device can extract all data it supports
// - till now only colours are implemented
//
class TTexture : public TMultiRef
{
protected:
    // actual colour of the object
    TColour  _colour;

    // ambient, diffuse reflection coeff
    real     _ambient, _diffuse;

    // specular coefficient
    real     _specular, _spec_degree;

    // transmissive coefficient
    real     _transmiss;
    
    // index of refraction
    real     _ior;
    
public:
    //////////////////////////////////////////////
    //
    // constructor and destructor
    //

    TTexture ()
            : _colour(0,0,0), _ambient(0.0), _diffuse(0.0),
              _specular(0.0), _spec_degree( 1 ),
              _transmiss(0.0), _ior(1.0)
    {}
    
    TTexture ( const TColour & c )
            : _colour(c), _ambient(0.0), _diffuse(0.0), _specular(0.0),
              _spec_degree( 1 ), _transmiss(0.0), _ior(1.0)
    {}
    
    TTexture ( const TTexture & t ) { *this = t; }
    
    virtual ~TTexture () {}

    //////////////////////////////////////////////
    //
    // return colour at given point
    //

    virtual void colour ( const TVector & v, TColour & col ) const { col = _colour; }

    //////////////////////////////////////////////
    //
    // access the local data
    //

    // get/set colour
    const TColour & colour () const { return _colour; }
    void set_colour ( const TColour & c ) { _colour = c; }

    //
    // get coefficients
    //
    
    real ambient   () const { return _ambient; }
    real diffuse   () const { return _diffuse; }
    real specular  () const { return _specular; }
    real specular_degree () const { return _spec_degree; }
    real transmiss () const { return _transmiss; }
    real ior       () const { return _ior; }
    
    //
    // set coefficients
    //
    
    void set_ambient  ( real amb  ) { _ambient  = min_max(  amb, 0.0, 1.0 ); }
    void set_diffuse  ( real diff ) { _diffuse  = min_max( diff, 0.0, 1.0 ); }
    
    void set_specular ( real spec, real degree )
    { _specular = min_max( spec, 0.0, 1.0 ); _spec_degree = degree; } 

    void set_transmiss  ( real t ) { _transmiss = t; }
    void set_ior        ( real f ) { _ior = (f >= 1.0 ? f : 1.0); }
    
    //////////////////////////////////////////////
    //
    // misc. methods
    //

    TTexture & operator = ( const TColour & c ) { _colour = c; return *this; }
    TTexture & operator = ( const TTexture & t )
    {
        _colour      = t._colour;
        _ambient     = t._ambient;
        _diffuse     = t._diffuse;
        _specular    = t._specular;
        _spec_degree = t._spec_degree;
        _transmiss   = t._transmiss;
        _ior         = t._ior;
        
        return *this;
    }

    // return copy of texture
    TTexture * copy () const { return new TTexture( *this ); }
};

} // namespace

#endif  // __TTEXTURE_HH
