//
// Project : lib3D
// File    : TPointLight.cc
// Purpose : class for a point-light (undirectional)
// Author  : Ronald Kriemann
//
// arch-tag: b2f3a3d9-2fae-4d29-a2c8-5f1170ce4631
//

#include "TPointLight.hh"

namespace RGL
{

// constants in the light-source-attenuation-term
#define C1      0.0
#define C2      1.0
#define C3      0.0
    
void
TPointLight::illuminate ( const THPDesc & hp,
                          const TVector & origin,
                          TColour       & colour ) const
{
    TVector  l;
    real     f_att, depth;
    real     f_diff = 0.0;
    real     f_spec = 0.0;

    l.set( 1.0, _pos, -1.0, hp.hitpoint() );
    depth = l.normalise();
    
    // apply light-source-attenuation-term
    f_att = 1.0 / (C1 + depth * ( C2 + C3 * depth ));
    if ( f_att > 1.0 )
        f_att = 1.0;
    
    // compute diffuse reflection coefficient
    f_diff = comp_diffuse( hp.object(), hp.normal(), l );

    // compute specular reflection coefficient
    TVector  v;

    v.set( 1.0, origin, -1.0, hp.hitpoint() );

    v.normalise();
    f_spec = comp_specular( hp.object(), hp.normal(), l, v );

    // set colour
    TColour  col;

    hp.object()->texture()->colour( hp.hitpoint(), col );
    
    colour.set_rgb( _colour.red()   * ((col.red()   * f_diff) + f_spec),
                    _colour.green() * ((col.green() * f_diff) + f_spec),
                    _colour.blue()  * ((col.blue()  * f_diff) + f_spec) );
}

//
// test if light is visible from given point
//
bool
TPointLight::is_visible ( const TVector    & point,
                          const TIntersect & intersect ) const
{
    real  distance;
    TRay  ray;

    ray.origin() = point;
    ray.direction().set( 1.0, _pos, -1.0, point );

    //
    // first check last object
    //

    if ( _last_obj != NULL )
    {
        if (_last_obj->intersect( ray, distance ))
        {
            // check if intersection is between light and given point
            if ((distance > EPSILON) && (distance <= 1.0))
                return false;
        }// if
        else
            _last_obj = NULL;
    }// if
    
    //
    // look if ray intersects with an object
    //

    TGraphObject * obj;

    obj = intersect.intersect( ray, distance, EPSILON, 1.0 );
    
    if (obj != NULL)
    {
        _last_obj = obj;
        return false;
    }// if
    
    // no object in the way
    return true;
}

} // namespace

