//
// Project : lib3D
// File    : TSpotLight.cc
// Purpose : class for a point-light (undirectional)
// Author  : Ronald Kriemann
//
// arch-tag: 021809db-9329-4c40-9a2c-66e57a7369f3
//

#include <cmath>

#include "TSpotLight.hh"

namespace RGL
{

// constants in the light-source-attenuation-term
#define C1  0.1
#define C2  0.1
#define C3  0.1

//////////////////////////////////////////////
//
// constructor and destructor
//

TSpotLight::TSpotLight ( const TVector & pos,
                         const TVector & look_at,
                         real            angle,
                         const TColour & col )
        : _pos( pos ), _colour( col ), _last_obj( NULL )
{
    _normal.set( 1.0, look_at, -1.0, _pos ).normalise();
    _max_angle = cos( angle * M_PI / 180.0 );
}

//////////////////////////////////////////////
//
// illumination methods
//

void
TSpotLight::illuminate ( const THPDesc & hp,
                         const TVector & origin,
                         TColour       & colour ) const
{
    TVector  l, r, v;
    real     angle;
    real     f_att, depth;
    real     f_diff = 0.0;
    real     f_spec = 0.0;
    real     f_spot = 0.0;
    real     d;
    
    l.set( 1.0, _pos, -1.0, hp.hitpoint() );
    depth = l.normalise();
    
    // apply light-source-attenuation-term
    f_att = 1.0 / (C1 + (C2 * depth) + (C3 * depth * depth));
    if ( f_att > 1.0 )
        f_att = 1.0;

    // compute diffuse reflection
    d = dot( hp.normal(), l );
    angle = d;
    if (angle > 0.0)
        f_diff = hp.object()->texture()->diffuse() * angle;
    
    // compute specular reflection
    r.set( 2.0 * d, hp.normal(), -1.0, l ).normalise();
    
    v.set( 1.0, origin, -1.0, hp.hitpoint() ).normalise();
    
    angle = dot( r, v );
    if ( angle > 0.0 )
    {
        real  degree = hp.object()->texture()->specular_degree();
        
        f_spec = hp.object()->texture()->specular();
        
        for ( uint i = 0; i < degree; i++ )
            f_spec *= angle;;
    }// if

    //
    // calculate spotlight
    //

    // compute vector from position to given point
/*    l.scale( -1 );

    angle = dot( l, _normal );
    if ( angle >= _max_angle )
    {
        f_spot = 1.0;
        for ( i = 0; i < 20; i++ )
            f_spot *= angle;;
            }// if */
    f_spot = 1.0;
    
    // set colour
    colour.set_rgb( _colour.red()   *
                   f_spot * ((hp.object()->texture()->colour().red()   * f_diff) + f_spec),
                   _colour.green() *
                   f_spot * ((hp.object()->texture()->colour().green() * f_diff) + f_spec),
                   _colour.blue()  *
                   f_spot * ((hp.object()->texture()->colour().blue()  * f_diff) + f_spec) );
}

//
// test if light is visible from given point
//
bool
TSpotLight::is_visible ( const TVector    & point,
                         const TIntersect & intersect ) const
{
    //
    // first check if point is in cone
    //

    TVector l;

    l.set( 1.0, point, -1.0, _pos ).normalise();

    if ( dot( l, _normal ) < _max_angle )
        return false;
    
    //
    // then check intersections with objects
    //

    TRay   ray;
    real   distance;

    ray.origin() = point;
    ray.direction().set( 1.0, _pos, -1.0, point );
    
    //
    // first check last object
    //

    if ( _last_obj != NULL )
    {
        if (_last_obj->intersect( ray, distance ))
        {
            // check if intersection is between light and given point
            if ((distance > EPSILON) && (distance <= 1.0))
                return false;
        }// if
        else
            _last_obj = NULL;
    }// if
    
    //
    // look if ray intersects with an object
    //

    TGraphObject * obj;

    obj = intersect.intersect( ray, distance, EPSILON, 1.0 );
    
    if (obj != NULL)
    {
        _last_obj = obj;
        return false;
    }// if
    
    // no object in the way
    return true;
}

//////////////////////////////////////////////
//
// virtual constructors
//

TSpotLight *
TSpotLight::clone () const
{
    TVector  v;

    v.set( 1.0, _pos, 1.0, _normal );
    
    return new TSpotLight( _pos, v, _max_angle, _colour ); }

} // namespace

