//
// Project : prender
// File    : main.cc
// Purpose : main program
// Author  : Ronald Kriemann
//
// arch-tag: 64c70319-9d96-4fd6-b878-961b7da5db6b
//

#if defined (SUNOS)
#include <unistd.h>
#else
#include <getopt.h>
#endif

#include "TTimer.hh"
#include "TFile.hh"
#include "TImage.hh"
#include "TXImage.hh"
#include "TRaytracer.hh"
#include "TParallelRT.hh"
#include "TSuperSamplingRT.hh"
#include "TAdaSuperSampRT.hh"
#include "TAdaptiveRT.hh"
#include "TImageIO.hh"
#include "TParser.hh"
#include "TPinholeCamera.hh"
#include "TOrthoCamera.hh"
#include "TThinLensCamera.hh"

using namespace RGL;

/////////////////////////////////////////////////////
//
// local data
//
/////////////////////////////////////////////////////

const std::string VERSION = "v0.9.1";

// different camera types
enum { CAMERA_PINHOLE,
       CAMERA_ORTHO,
       CAMERA_THINLENS };

/////////////////////////////////////////////////////
//
// function forwards
//
/////////////////////////////////////////////////////

void usage     ( const char * msg, int stat );
void set_bench ( TWorld * world, TCamera * & camera );
void read_ply  ( TWorld * world, const char * filename, TProgressBar * progress );

/////////////////////////////////////////////////////
//
// local variables
//
/////////////////////////////////////////////////////

#if defined(LINUX) || defined(DARWIN)
static struct option longopts[] = {
    { "benchmark",  no_argument,       0, 'b' },
    { "contrast",   required_argument, 0, 'c' },
    { "depth",      required_argument, 0, 'd' },
    { "gamma",      required_argument, 0, 'g' },
    { "help",       no_argument,       0, 'h' },
    { "outfile",    required_argument, 0, 'o' },
    { "nprocs",     required_argument, 0, 'p' },
    { "preview",    no_argument,       0, 'P' },
    { "resolution", required_argument, 0, 'r' },
    { "sampling",   required_argument, 0, 's' },
    { "scene",      required_argument, 0, 'f' },
    { "tilesize",   required_argument, 0, 't' },
    { "verbose",    required_argument, 0, 'v' },
    { 0,          0,                 0,  0  }
};
#endif

/////////////////////////////////////////////////////
//
// main routine
//
/////////////////////////////////////////////////////

int
main ( int argc, char ** argv )
{
    uint     x = 600, y = 400;
    TString  filename  = "scene/scene.wd";
    TString  outfile   = "scene/scene.png";
    bool     out_set   = false;
    uint     depth     = 10;
    uint     sampling  = 1;
    real     contrast  = 0.0;
    uint     nprocs    = 1;
    uint     tile_size = 32;
    real     gamma     = 2.2;
    bool     benchmark = false;
    bool     preview   = false;

    set_verbosity( 1 );

    //
    // parse command line
    //

    int opt;
    
#if defined(LINUX) || defined(DARWIN)
    while (( opt = getopt_long( argc, argv, "bc:d:f:g:ho:p:Pr:s:t:v:", longopts, NULL )) != -1 )
#else
    while (( opt = getopt( argc, argv, "bc:d:f:g:ho:p:Pr:s:t:v:" )) != -1 )
#endif
    {
        switch ( opt )
        {
        case 'b' : benchmark = true; break;
        case 'c' : contrast  = atof( optarg ); break;
        case 'd' : depth     = atoi( optarg ); break;
        case 'f' :
            filename = optarg;
            
            if ( ! out_set )
            {
                for ( int j = filename.length()-1; j >= 0; j-- )
                {
                    if ( filename[j] == '.' )
                    {
                        outfile  = filename.substr( 0, j );
                        outfile += ".png";
                        break;
                    }// if
                }// for
            }// if
            break;

        case 'g' : gamma = max( 1.0, min( 3.0, atof( optarg ) ) ); break;
        case 'h' : usage( "", 0 ); break;
        case 'o' : outfile = optarg; break;
        case 'p' : nprocs = atoi( optarg ); break;
        case 'P' : preview = true; break;
        case 'r' :
        {
            TString            tmp = optarg;
            TArray< TString >  entries;

            tmp.split( "x", entries );

            if ( entries.size() != 2 )
                usage( "no valid geometry definition after -r", 1 );
                
            x = atoi( entries[0].c_str() );
            y = atoi( entries[1].c_str() );
                    
            if (( x == 0 ) || ( y == 0 ))
                usage( "no valid geometry definition after -r", 1 );
        }
        break;

        case 's' : sampling = atoi( optarg ); break;
        case 't' : tile_size = atoi( optarg ); break;
        case 'v' : set_verbosity( atoi( optarg ) ); break;

        case ':' : usage( "no parameter supplied (see -h)", 1 );

        default  :
        case '?' : usage( "unknown option (see -h)", 1 );
        }// switch 
    }// while

    //
    // render the world
    //

    TRaytracer   * rt    = NULL;
    TWorld       * world = new TWorld;
    TCamera      * camera;
    TProgressBar * progress = (verbose(2) ? new TProgressBar : NULL);

    if ( benchmark )
    {
        x         = 1800;
        y         = 1200;
//         x         = 750;
//         y         = 500;
        sampling  = 1;
        depth     = 10;
        contrast  = 0.0;
        tile_size = 32;
        filename  = "<internal>";
        outfile   = "benchmark.png";

        set_bench( world, camera );
    }// if
    else
    {
        TParser   parser( world );
        TFile   * file     = file_fac.open( filename, "r" );

        camera = new TPinholeCamera;
        
        parser.parse( file );
        parser.apply_camera( camera );
        
        delete file;
    }// else
    
    // read_ply( world, "/home/rok/download/models/dinosaur.ply.bz2" );
    // add_objects( world );
    
    if ( verbose(1) )
    {
        std::cout << "(prender) options : " << std::endl
                  << "  input/output file = " << filename << " / " << outfile << std::endl
                  << "  objects in world  = " << world->objects().size() << std::endl
                  << "  image resolution  = " << x << " x " << y << ",\t"
                  << "  gamma correction = " << gamma << std::endl
                  << "  super sampling    = " << sampling << ",\t\t"
                  << "  max. contrast    = " << contrast << std::endl
                  << "  processors to use = " << nprocs << ",\t\t"
                  << "  size of tiles    = " << tile_size << std::endl;
    }// if

    if ( contrast > 0.0 )
    {
        if ( sampling > 1 )
        {
            TAdaSuperSampRT  * ssrt = new TAdaSuperSampRT;

            ssrt->set_super_sampling( sampling );
            ssrt->set_contrast( contrast );

            rt = ssrt;
        }// if
        else
        {
            TAdaptiveRT * art = new TAdaptiveRT;
            
            art->set_max_contrast( contrast );
            
            rt = art;
        }// else
    }// if
    else
    {
        TSuperSamplingRT  * ssrt = new TSuperSamplingRT;
        
        ssrt->set_super_sampling( sampling );
        
        rt = ssrt;
    }// else

    if ( nprocs > 1 )
        rt = new TParallelRT( rt, nprocs, tile_size );
    
    rt->set_background( TColour( "black" ) );
    rt->set_camera( camera );
    rt->set_world( world );
    rt->set_max_depth( depth );

    //
    // finally render the image
    //

    TImage * image = (preview ? new TXImage( x, y, gamma ) : new TImage( x, y ));
    TTimer   timer( RGL::REAL_TIME );

    timer.start();
    
    rt->render_world( * image, progress );
    
    timer.pause();
    
    if ( verbose(1) )
    {
        //
        // print statistics
        //
        
        std::cout << "(prender) statistics :" << std::endl
                  << "  total no. of rays   = " << rt->no_of_rays() << std::endl
                  << "  reflected/refracted = " << rt->reflected_rays()
                  << " / " << rt->refracted_rays() << std::endl
                  << "  max level           = " << rt->deepest_level() << std::endl
                  << "  time to raytrace    = " << timer << std::endl
                  << "    per ray / pixel   = ";

        if ( rt->no_of_rays() != 0 )
            std::cout << to_str( "%.2f", 1e6 * timer.elapsed() / real(rt->no_of_rays()) );
        else
            std::cout << "0.0";

        std::cout << " us / ";
        
        if ( x*y != 0 )
            std::cout << to_str( "%.2f", 1e6 * timer.elapsed() / real(x*y) );
        else
            std::cout << "0.0";

        std::cout << " us" << std::endl;
    }// if

    delete progress;
    delete rt;

    if ( preview )
        ((TXImage *) image)->end();
    
    //
    // print image
    //
    
    TImgIOHandler  img_io( gamma );

    img_io.save( outfile, * image );
    
    return 0;
}

/////////////////////////////////////////////////////
//
// local functions (implementation)
//
/////////////////////////////////////////////////////

//
// print usage information
//
void
usage ( const char * msg, int stat )
{
    if ( stat != 0 )
    {
        std::cout << "error: " << msg << std::endl
                  << "try \"-h\" for complete option list" << std::endl;
        exit( stat );
    }// if

    std::cout
        << "prender " << VERSION << " by Ronald Kriemann" << std::endl
        << "usage : prender [ options ]" << std::endl
        << "where options include :" << std::endl
        << "  -b, --benchmark      : do a benchmark with a fixed scene/resolution" << std::endl
        << "  -c, --contrast <num> : contrast value for adaptive sampling (default = 0.0)" << std::endl
        << "  -d, --depth <int>    : maximal recursion depth (default = 10)" << std::endl
        << "  -f, --scene <file>   : set scene description (default = \"scene.wd\")" << std::endl
        << "  -g, --gamma <num>    : set gamma correction (default = 2.2, 1.0 ... 3.0)" << std::endl
        << "  -h, --help           : print this help message" << std::endl
        << "  -o, --outfile <file> : set output image-name (default = \"scene.png\")" << std::endl
        << "  -p, --procs <int>    : set no. of processors to use (default = 1)" << std::endl
        << "  -P, --preview        : show rendering of image (default = off)" << std::endl
        << "  -r, --resolution wxh : set size of image (default = 400x400)" << std::endl
        << "  -s, --samples <int>  : set no. of samples per pixel (default = 1)" << std::endl
        << "  -t, --tilesize <int> : size of tiles in parallel mode (default = 32)" << std::endl
        << "  -v, --verbose <int>  : set verbosity level (default = 1, 0 = off)" << std::endl;

    exit( 0 );
}

#include "TPlane.hh"
#include "TSphere.hh"
#include "TTriangle.hh"
#include "TAmbient.hh"
#include "TPointLight.hh"
#include "TChecker.hh"
#include "TMarble.hh"
#include "TVRML.hh"

//
// put objects into the world
//
void
sphere ( TWorld * world, TTexture * tex,
         const TVector & c, const TVector & dir, real rad,
         uint lvl, uint max_lvl )
{
    TSphere * s = new TSphere( c, rad );

    s->set_texture( tex );
    
    world->add( s );

    if ( lvl >= max_lvl )
        return;

    //
    // randomly choose point on sphere for direction of
    // subsphere, but keep distances over a defined threshold
    //

    TVector  axis( 1, 0, -1 );
    TVector  dirs[3];
    TMatrix  rot, rot2;
    real     d = 1.0 / std::sqrt( 2.0 );
    real     r = rad / 3.0;
    TRNG     rng( 1 );

    axis.normalise();
    rot.set_rotate( axis, std::asin( 2.0 / std::sqrt( 6.0 ) ) );

    rot.mul_vec( TVector( d, d, 0 ), dirs[0] );
    rot.mul_vec( TVector( d, 0, -d ), dirs[1] );
    rot.mul_vec( TVector( 0, d, -d ), dirs[2] );

    for ( uint i = 0; i < 3; ++i )
    {
        rot.set_rotate( TVector( 0, 1, 0 ), real(i) * 2.0 * M_PI / 3.0 );
        
        for ( uint j = 0; j < 3; ++j )
        {
            TVector  t, cdir;

            rot.mul_vec( dirs[j], t );

            //
            // rotate new sphere into right direction
            //
            
            if      ( dir.y() >=  1.0 ) rot2.set_identity();
            else if ( dir.y() <= -1.0 ) rot2.set_rotate( VEC_Z, M_PI);
            else
            {
                axis.set_cross( VEC_Y, dir ).normalise();
                rot2.set_rotate( axis, std::acos( dot( VEC_Y, dir ) ) );
            }// else

            rot2.mul_vec( t, t );
            t.scale( rad + r ).add( 1.0, c );
            cdir.set( 1.0, t, -1.0, c ).normalise();
            sphere( world, tex, t, cdir, r, lvl+1, max_lvl );
        }// for
    }// for
}

void
set_bench ( TWorld * world, TCamera * & camera )
{
    TGraphObject * obj = NULL;

    obj = new TPlane( TVector( 0,  0,  0 ),
                      TVector( 0,  0,  1 ),
                      TVector( 1,  0,  0 ) );
    obj->set_texture( new TChecker( TColour( "black" ),
                                    TColour( "white" ),
                                    TVector( 2, 0, 2 ),
                                    false ) );
//    obj->set_texture( new TMarble( 0.4, 10, 8, 2 ) );
    obj->texture()->set_ambient( 0.3 );
    obj->texture()->set_diffuse( 0.6 );
    obj->texture()->set_specular( 0.2, 300 );
    world->add( obj );

#if 0
    obj = new TSphere( TVector( 0, 1.35, 0 ), 0.15 );
    obj->set_texture( new TMarble( 1, 10, 8, 2 ) );
    obj->texture()->set_ambient( 0.3 );
    obj->texture()->set_diffuse( 0.6 );
    obj->texture()->set_specular( 0.05, 64 );
    world->add( obj );
    
    TTexture * tex = new TTexture;
    
    tex->set_colour( TColour( "PeachPuff" ) );
    tex->set_ambient( 0.2 );
    tex->set_diffuse( 0.7 );
    tex->set_specular( 0.2, 32 );
    
    for ( real phi = 0, y = 0.1, r = 0.5, d = 0.0;
          phi < 3*M_PI;
          phi += 0.1, y += 0.0075+d, r -= 0.004, d += 0.00008 )
        for ( real psi = 0.0; psi < 2.0*M_PI; psi += M_PI_4 )
        {
            obj = new TSphere( TVector( r*std::sin(phi+psi),
                                        y,
                                        r*std::cos(phi+psi) ),
                               0.0225 );
            obj->set_texture( tex );
            world->add( obj );
        }// for
    
    world->add_light( new TAmbient( TColour( "gray30" ) ) );
    world->add_light( new TPointLight( TVector( 0,  7, -2 ), TColour( "gray90" ) ) );

    camera = new TPinholeCamera;
    camera->set_origin_dir( TVector(  -3,  1.5, -5 ),
                            TVector(  0.85*1.9, 0.85*-0.5, 0.85*3 ) );
#else
    TTexture * tex = new TTexture;
    
    tex->set_colour( TColour( "red" ) );
    tex->set_ambient( 0.2 );
    tex->set_diffuse( 0.5 );
    tex->set_specular( 0.8, 64 );

    for ( real z = 0; z < 10; z += 1.0 )
    {
        obj = new TSphere( TVector( -0.5, 0.5, z ), 0.2 );
        obj->set_texture( tex );
        world->add( obj );
    }// for

    tex = new TTexture;
    
    tex->set_colour( TColour( "gold" ) );
    tex->set_ambient( 0.2 );
    tex->set_diffuse( 0.8 );
    tex->set_specular( 0.0, 64 );

    sphere( world, tex, TVector( 1, 0.6, 3 ), VEC_Y, 0.4, 0, 4 );
    
    world->add_light( new TAmbient( TColour( "gray30" ) ) );
    world->add_light( new TPointLight( TVector( 0,  7, -2 ), TColour( "gray90" ) ) );

    if ( false )
    {
        camera = new TPinholeCamera;
        camera->set_film_dist( 150.0 / 1000.0 );
        camera->set_origin_lookat( TVector( -2.0, 3, -10 ), TVector( -0.1, 0.85, 1 ) );
    }// if

    if ( true )
    {
        TThinLensCamera * tlcam = new TThinLensCamera;
        
        tlcam->set_film_dist( 150.0 / 1000.0 );
        tlcam->set_origin_lookat( TVector( -2.0, 3, -10 ), TVector( -0.1, 0.85, 1 ) );
        tlcam->set_thin_lens( 2.8, 11, 20 );

        camera = tlcam;
    }// if

    if ( false )
    {
        camera = new TOrthoCamera;
        camera->set_film_size( 1.2*3, 1.2*2 );
        camera->set_origin_lookat( TVector( -2.0, 3, -10 ), TVector( -0.25, 1.1, 1 ) );
    }// if
#endif
}

//
// read model defined by ply-file
//
void
read_ply ( TWorld * world, const char * filename, TProgressBar * progress )
{
    TFile * file = file_fac.open( filename, "r" );

    if ( file == NULL )
        return;

    //
    // read basic description of model
    //

    TString            buf( 256 );
    uint               n_vertices = 0, n_faces = 0;
    TArray< TString >  substr;

    file->gets( buf );
    buf[ buf.length() - 1 ] = '\0';
    if ( buf != "ply" )
    {
        std::cout << "(read_ply) file is not in ply-format" << std::endl;
        file->close();
        return;
    }// if

    // read complete header
    do
    {
        if ( file->gets( buf ) != 0 )
            break;

        if ( buf.split( " \t\n", substr ) >= 3 )
        {
            if (( substr[0] == "element" ) &&
                ( substr[1] == "vertex" ) )
                n_vertices = atoi( substr[2].c_str() );
            else if (( substr[0] == "element" ) &&
                     ( substr[1] == "face" ) )
                n_faces = atoi( substr[2].c_str() );
        }// if
    } while (( substr.size() > 0 ) && ( substr[0] != "end_header" ));

    //
    // read vertices
    //

    real              x, y, z;
    TArray< TVector > vertices( n_vertices ); 

    if ( progress != NULL )
    {
        progress->set_prefix( "reading vertices :" );
        progress->init( 0, n_vertices, 0 );
    }// if
    
    for ( uint i = 0; i < n_vertices; i++ )
    {
        if ( file->gets( buf ) != 0 )
            break;

        if ( progress != NULL )
            progress->advance( 1 );
        
        if ( buf.split( " \t\n", substr ) >= 3 )
        {
            x = 10.0 * atof( substr[0].c_str() );
            y = 10.0 * atof( substr[1].c_str() );
            z = 10.0 * atof( substr[2].c_str() );

            vertices[i].set ( x, y, z );
        }// if
    }// for

    //
    // read facess and create objects
    //
    
    TTexture        * tex = new TTexture;
    TGraphObject    * obj;
    uint              mod;
    int               i1, i2, i3;
    TVector           center, v;
    real              radius, f;
    
    tex->set_colour( TColour( "gold" ) );
    tex->set_ambient( 0.3 );
    tex->set_diffuse( 0.6 );

    mod = n_faces / 2000000;
    if ( mod == 0 )
        mod = 1;

    if ( progress != NULL )
    {
        progress->set_prefix( "reading faces :" );
        progress->init( 0, n_faces, 0 );
    }// if
    
    for ( uint i = 0; i < n_faces; i++ )
    {
        if ( file->gets( buf ) != 0 )
            break;

        if ( progress != NULL )
            progress->advance( 1 );
        
        if ( i % mod != 0 )
            continue;
        
        if ( buf.split( " \t\n", substr ) >= 3 )
        {
            i1 = atoi( substr[1].c_str() );
            i2 = atoi( substr[2].c_str() );
            i3 = atoi( substr[3].c_str() );

            // compute center and radius of face
#if 1
            center.set( 1.0, vertices[i1] );
            center.add( 1.0, vertices[i2], 1.0, vertices[i3] );
            center.scale( 1.0 / 3.0 );

            radius = 0.0;
            f = v.set( 1.0, center, -1.0, vertices[i1] ).norm2();
            radius = max( radius, f );
            f = v.set( 1.0, center, -1.0, vertices[i2] ).norm2();
            radius = max( radius, f );
            f = v.set( 1.0, center, -1.0, vertices[i3] ).norm2();
            radius = max( radius, f );

            obj = new TSphere( center, 2.0*radius );
#else
            obj = new TTriangle( vertices[i1], vertices[i2], vertices[i3] );
#endif
            obj->set_texture( tex );
            world->add( obj );
        }// if
    }// for

    if ( progress != NULL )
        progress->finish();

    file->close();
    delete file;
}

