//
// Project : lib3D
// File    : TCamera.cc
// Purpose : Declaration of a class for central-projection
// Author  : Ronald Kriemann
//
// arch-tag: f099aa79-dfc5-43fe-b95c-e0d44925cf61
//

#include "TRNG.hh"
#include "TCamera.hh"

namespace RGL
{

//////////////////////////////////////////////
//
// constructor and destructor
//

TCamera::TCamera ()
{
    // default ist at (0,0,-1) looking at (0,0,0)
    _origin = TVector( 0, 0, -1 );
    _lookat = TVector( 0, 0,  0 );

    _back   = -1;

    _film_width  = 36.0 / 1000.0;  // 36 mm
    _film_height = 24.0 / 1000.0;  // 24 mm
    _film_dist   = 50.0 / 1000.0;  // 50 mm
    
    _width  = 1;
    _height = 1;

    construct_by_dir( TVector( 0, 0, 1 ) );
}

TCamera::~TCamera ()
{
}

//////////////////////////////////////////////
//
// define view plane
//

//
// construct plane by a and b such that u || a
// and v lies in the plane of a and b
//
void
TCamera::construct_by_plane ( const TVector & a,
                              const TVector & b )
{
    real  len_a = a.norm2();
    real  len_b = b.norm2();
    
    if (( len_a < EPSILON ) || ( len_b < EPSILON ))
    {
        std::cout << "(TCamera) construct_by_plane : given vector a has length 0" << std::endl;
        return;
    }// if

    _width  = len_a;
    _height = len_b;
    
    _u.set( 1.0 / len_a, a );
    _w.set_cross( a, b ).normalise();
    _v.set_cross( _w, _u );

    update();
}

//
// construct view plane by a given direction and
// orthogonal but otherwise arbitrary plane vectors
//
void
TCamera::construct_by_dir ( const TVector & a )
{
    real  len_a = a.norm2();
    
    if ( len_a < EPSILON )
    {
        std::cout << "(TCamera) construct_by_dir : given vector has length 0" << std::endl;
        return;
    }// if

    _w.set( 1.0 / len_a, a );
    _u.set_cross( VEC_Y, _w );

    // in case, that the direction is too parallel to x, try y
    if ( _u.norm2() < EPSILON )
        _u.set_cross( VEC_X, _w );
    
    _v.set_cross( _w, _u );

    update();
}

//
// define origin and lookat of camera
//
void
TCamera::set_origin_lookat ( const TVector & orig, const TVector & lookat )
{
    TVector  t;

    t.set( 1.0, lookat, -1.0, orig );
    
    if ( t.norm2() < EPSILON )
    {
        std::cout << "(TCamera) set_origin_lookat : |lookat - origin| = 0" << std::endl;
        return;
    }// if
    
    _origin = orig;
    _lookat = lookat;

    construct_by_dir( t );
}

//
// define origin and direction of camera
//
void
TCamera::set_origin_dir ( const TVector & orig, const TVector & dir )
{
    TVector  la;

    la.set( 1.0, orig, 1.0, dir );

    set_origin_lookat( orig, la );
}

//
// copy data from other camera
//
void
TCamera::set ( const TCamera & cam )
{
    _origin      = cam._origin;
    _lookat      = cam._lookat;
    _u           = cam._u;
    _v           = cam._v;
    _w           = cam._w;
    _ulc         = cam._ulc;
    _vp_right    = cam._vp_right;
    _vp_down     = cam._vp_down;
    _width       = cam._width;
    _height      = cam._height;
    _front       = cam._front;
    _back        = cam._back;
    _film_width  = cam._film_width;
    _film_height = cam._film_height;
    _film_dist   = cam._film_dist;
}

//
// set dimensions of film (and hence of view plane)
//
void
TCamera::set_film_size ( real width, real height )
{
    if (( width <= 0.0 ) || ( height <= 0.0 ))
    {
        std::cout << "(TCamera) set_film_size : invalid size" << std::endl;
        return;
    }// if

    _film_width  = width;
    _film_height = height;

    update();
}

//
// set film distance from origin
//
void
TCamera::set_film_dist ( real dist )
{
    if ( dist <= 0.0 )
    {
        std::cout << "(TCamera) set_film_dist : invalid distance" << std::endl;
        return;
    }// if

    _film_dist = dist;
    
    update();
}
    
//////////////////////////////////////////////
//
// transform camera
//

void
TCamera::move ( const TVector & dist )
{
    _origin.add( 1.0, dist );
    _lookat.add( 1.0, dist );
    update();
}

void
TCamera::move_horiz ( real f )
{
    TVector  v;

    v.set( f, _u );
    move( v );
}

void
TCamera::move_vert  ( real f )
{
    TVector  v;

    v.set( f, _v );
    move( v );
}

void
TCamera::move_dir   ( real f )
{
    TVector  v;

    v.set( -f, _w );
    move( v );
}

void
TCamera::rotate ( const TVector & dir, real angle )
{
    rotate( VEC_0, dir, angle );
}

void
TCamera::rotate ( const TVector & orig,
                  const TVector & dir,
                  real angle )
{
    //
    // this methode stems from the Mesa-Lib which implements
    // a really compact and fast rotation-method
    //
    real  mag, s, c;
    real  x, y, z;
    real  xx, yy, zz, xy, yz, zx, xs, ys, zs, one_c;
    
    s = std::sin( angle );
    c = std::cos( angle );

    x = dir.x();
    y = dir.y();    
    z = dir.z();

    mag = norm2( dir );

    // no "real" direction given
    if (mag == 0.0)
        return;
    
    x /= mag;
    y /= mag;
    z /= mag;
    
    xx = x * x;  yy = y * y;  zz = z * z;
    xy = x * y;  yz = y * z;  zx = z * x;
    xs = x * s;  ys = y * s;  zs = z * s;
    
    one_c = 1.0 - c;

    TMatrix   m;
    TVector   up( VEC_Y );
    
    m(0,0) = (one_c * xx) + c;   m(0,1) = (one_c * xy) - zs;  m(0,2) = (one_c * zx) + ys;
    m(1,0) = (one_c * xy) + zs;  m(1,1) = (one_c * yy) + c;   m(1,2) = (one_c * yz) - xs;
    m(2,0) = (one_c * zx) - ys;  m(2,1) = (one_c * yz) + xs;  m(2,2) = (one_c * zz) + c;

    _origin.add( -1.0, orig );
    _lookat.add( -1.0, orig );
    up.add( -1.0, orig );
    
    m.mul_vec( _origin );
    m.mul_vec( _lookat );
    m.mul_vec( up );

    _origin.add( 1.0, orig );
    _lookat.add( 1.0, orig );
    up.add( 1.0, orig );
    
    update();
}

void
TCamera::rotate_horiz ( real angle )
{
    rotate( VEC_0, _v, angle );
}

void
TCamera::rotate_vert  ( real angle )
{
    rotate( VEC_0, _u, angle );
}

//
// update internal data
//
void
TCamera::update ()
{
    //
    // width and height are equal to film size
    // since ray direction is normalised, front is 1
    //

    _width  = _film_width;
    _height = _film_height;
    _front  = 1.0;
    
    //
    // set upper left corner, right and down vector of view plane
    //
    
    _ulc.set( 1.0, _origin, _film_dist, _w ).add( -_width/2.0, _u, _height/2.0, _v );
    _vp_right.set( _width, _u );
    _vp_down.set( -_height, _v );
}

///////////////////////////////////////////////////////
//
// output
//

//
// stream output
//
std::ostream &
operator << ( std::ostream & os, const TCamera & c )
{
    os << c.origin() << " -> " << c.look_at()
       << ", "
       << c.vp_width() << " * " << c.right()
       << "  x  "
       << c.vp_height() << " * " << c.up();

    return os;
}
    
} // namespace

