//
// Project : misc
// File    : TMemory.cc
// Purpose : class for gathering information about memory
// Author  : Ronald Kriemann
//
// arch-tag: 6763e638-bb02-4e0d-afb9-dbe5ff96bc2f
//

#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/resource.h>

#if defined(TRU64) || defined(AIX)
#include <sys/procfs.h>
#endif

#if defined(HPUX)
#include <sys/param.h>
#include <sys/pstat.h>
#endif

#include "ufo.hh"

#include "TMemory.hh"

namespace RGL
{

// global object
TMemory  MEMORY;

//
// return used memory (in kbytes)
//
unsigned long
TMemory::used () const
{
    unsigned long   mem_size = 0;

#if defined(LINUX)
    
    //
    // read size as first number in "/proc/<pid>/statm"
    //
    
    char           buf[256];
    FILE         * file;
    unsigned int   pages;
        
    snprintf( buf, sizeof(buf), "/proc/%d/statm", (unsigned int) getpid() );
        
    if ((file = fopen( buf, "r" )) == NULL)
    {
        perror( "(TMemory) used : open" );
        return 0;
    }// if
        
    fgets( buf, sizeof(buf), file );
        
    fclose( file );
        
    sscanf( buf, "%u", & pages );
    mem_size = ((unsigned long) pages) * ((unsigned long) getpagesize());
        
#elif defined(SUNOS)
#  if defined (__sparc_v9__) || defined (__sparcv9)
    
    //
    // read size of file "/proc/<pid>/as"
    //
    struct stat64  stat_buf;
    char           buf[256];
        
    snprintf( buf, sizeof(buf), "/proc/%d/as", uint(getpid()) );

    if (stat64( buf, & stat_buf ) < 0)
    {
        perror( "(TMemory) used : stat" );
        return 0;
    }// if
        
    mem_size = stat_buf.st_size;

#  else

    //
    // read size of file "/proc/<pid>/as"
    //
    struct stat  stat_buf;
    char         buf[256];
        
    snprintf( buf, sizeof(buf), "/proc/%d/as", uint(getpid()) );
        
    if (stat( buf, & stat_buf ) < 0)
    {
        perror( "(TMemory) used : stat" );
        return 0;
    }// if
        
    mem_size = stat_buf.st_size;

#  endif
#elif defined(TRU64)
    
    //
    // read size in "/proc/<pid>"
    //
    
    char              buf[256];
    int               file;
    struct prpsinfo * prpsinfo;
    struct prpsinfo   arrayprpsinfo[32];
    
    prpsinfo = arrayprpsinfo;
    snprintf( buf, sizeof(buf), "/proc/%d", (unsigned int) getpid() );
    
    if ((file = open( buf, O_RDONLY, 0555 )) < 0)
    {
        perror( "(TMemory) used : open" );
        return 0;
    }// if
    
    if ((ioctl(file, PIOCPSINFO, prpsinfo)) == -1)
    {
        perror( "(TMemory) used : ioctl" );
        close( file );
        return 0;
    }// if
    
    close( file );
    mem_size = prpsinfo->pr_size * getpagesize();
    
#elif defined(AIX)

    //
    // read size from "/proc/<pid>/psinfo"
    // (returns only memory allocated via brk, not mmap)
    //
      
    char           buf[256];
    int            file;
    struct psinfo  psi;

    snprintf( buf, sizeof(buf), "/proc/%d/psinfo", (unsigned int) getpid() );

    if ((file = open( buf, O_RDONLY, 0555 )) < 0)
    {
        perror( "(TMemory) used : open" );
        return 0;
    }// if
    
    if ( read( file, & psi, sizeof(psi) ) < 0 )
    {
        perror( "(TMemory) used : read" );
        close( file );
        return 0;
    }// if
    
    close( file );

    mem_size = psi.pr_size;

#elif defined(HPUX)

    //
    // use pstat functions to gather info
    //

    struct pst_status pst;

    if ( pstat_getproc( & pst, sizeof(pst), 0, getpid() ) != -1 )
    {
        mem_size  = pst.pst_dsize + pst.pst_tsize + pst.pst_ssize + pst.pst_mmsize;
        mem_size *= getpagesize();
    }// if
    else
        perror( "TMemory::used () : pstat_getproc" );

#else

#define NO_MEMORY_STAT
    
    return 0;
    
#endif
        
    return mem_size - _base;
}

//
// stream output
//
std::ostream &
operator << ( std::ostream & os, const TMemory & mem )
{
#ifndef NO_MEMORY_STAT
    long  used = mem.used();

    os << byte_out( used );
#endif
    
    return os;
}

} // namespace

