//
// Project : PHI
// File    : TProgressBar.cc
// Purpose : class for showing the progress in various formats
// Author  : Ronald Kriemann
//
// arch-tag: db74e57d-3c6d-4f61-9d03-a96ce9786e96
//

#include <iostream>
#include <cmath>

#include "ufo.hh"

#include "TProgressBar.hh"

namespace RGL
{

////////////////////////////////////////////////
//
// defines
//
////////////////////////////////////////////////

// are colours initialised; have to do this by hand, since
// UFO colours are also initialised later
static bool init_colours = false;

// bar colours
static TString PREFIX_COL;
static TString BAR_COL;
static TString PERCENT_COL;
static TString ETA_COL;

////////////////////////////////////////////////
//
// local data
//
////////////////////////////////////////////////

//
// constructor and destructor
//

TProgressBar::TProgressBar ( real min, real max, real curr )
        : _min(min), _max(max), _curr(curr),
          _old_val(-1), _text_len(0)
{
    _timer.real();
    _prefix  = "";
    _is_init = false;

    if ( ! init_colours )
    {
        init_colours = true;
        
        PREFIX_COL  = BOLD.colour_str();
        BAR_COL     = "";
        PERCENT_COL = BOLD.colour_str() + BLUE.colour_str();
        ETA_COL     = BOLD.colour_str() + RED.colour_str();
    }// if
}

TProgressBar::~TProgressBar ()
{
}

//
// change status
//

//
// initialise status
//
void
TProgressBar::init ( real min, real max, real curr )
{
    _min  = min;
    _max  = max;
    _curr = curr;
    
    _old_val = -1;

    _text_len = 0;

    _timer.start();

    _is_init = true;
    
    print();
}

//
// advance progress
//
void
TProgressBar::advance ( real f )
{
    _curr += f;

    print();
}

//
// finish output
//
void
TProgressBar::finish ()
{
    std::cout << "\r" << std::flush;

    for ( uint i = 0; i <= _text_len; i++ )
        std::cout << ' ';

    std::cout << "\r" << std::flush;

    _is_init = false;
}

//
// print progress
//

#define MAX_LEN  40

void
TProgressBar::print ()
{
    real  percent = 100.0 * (_curr - _min) / (_max - _min);

    // check if update is really necessary
    if ( uint(percent) == uint(_old_val) )
        return;
    
    _old_val = percent;
    
    //
    // print actual bar
    //
    
    TString  output, str;
    uint     len = 0;
    uint     bar_len = uint(real(MAX_LEN) * percent / 100.0);

    output = '\r';

    if ( _prefix != "" )
    {
        output += PREFIX_COL + _prefix + NORMAL.colour_str();
        len = _prefix.length();
    }// if
    
    output += BAR_COL + '[';
    len++;

    for ( uint i = 0; i < MAX_LEN; i++ )
    {
        if ( i < bar_len ) output += '=';
        else               output += ' ';
    }// for
    
    len += MAX_LEN;

    output += ']' + NORMAL.colour_str();
    len++;

    //
    // print percentage
    //

    output += PERCENT_COL;

    str     = to_str( " %3d %%", uint(percent) );
    output += str;
    len    += str.length();

    output += NORMAL.colour_str();
    
    //
    // print ETA
    //

    if ( percent > 0.0 )
    {
        real ela   = _timer.elapsed();
        real eta_s = ela * 100.0 / percent - ela;
        real eta_m, eta_h;
        
        eta_s = max( eta_s, 0.0 );
        eta_m = std::floor( eta_s / 60.0 );
        eta_h = std::floor( eta_m / 60.0 );
        
        output += ETA_COL;
    
        if ( eta_h > 0.0 )
        {
            eta_s -= eta_m * 60.0;
            eta_m -= eta_h * 60.0;
            
            str = to_str( " ETA %d:%02d:%02d h ", uint(eta_h), uint(eta_m), uint(eta_s) );
        }// if
        else if ( eta_m > 0.0 )
        {
            eta_s -= eta_m * 60.0;
            str = to_str( " ETA %d:%02d m ", uint(eta_m), uint(eta_s) );
        }// if
        else
            str = to_str( " ETA %d s ", uint(eta_s) );

        output += str;
        len    += str.length();
        output += NORMAL.colour_str();
    }// if
    else
    {
        output += ' ';
        len++;
    }// else
        

    while ( len < _text_len )
    {
        output += ' ';
        len++;
    }// while
    
    std::cout << output << std::flush;

    _text_len = len;
}

}// namespace
