//
// Project : misc
// File    : TSobol.cc
// Purpose : class for a Sobol sequence generator
// Author  : Ronald Kriemann
//
// arch-tag: 642bdc75-5cdc-4649-a6ca-1bf8f06d6f45
//

#include "TSobol.hh"

////
////
//// Implementation stems from the GNU Scientific Library
//// (just restricted to 6 dimensions)
////
////

namespace RGL
{

///////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////
//
// local data
//
///////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////

//
// primitive polynomials in binary encoding
//
static const int primitive_polynomials[6] = { 1, 3, 7, 11, 13, 19 };

//
// degrees of the primitive polynomials
//
static const int degree_table[6] = { 0, 1, 2, 3, 3, 4 };

//
// initial values for direction tables
//
static const int v_init[8][6] =
{
  { 0, 1, 1, 1, 1, 1 },
  { 0, 0, 1, 3, 1, 3 }, 
  { 0, 0, 0, 7, 5, 1 }, 
  { 0, 0, 0, 0, 0, 1 }, 
  { 0, 0, 0, 0, 0, 0 }, 
  { 0, 0, 0, 0, 0, 0 }, 
  { 0, 0, 0, 0, 0, 0 }, 
  { 0, 0, 0, 0, 0, 0 }
};

///////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////
//
// TSobol (implementation)
//
///////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////

////////////////////////////////////////////
//
// access Sobol sequence
//

//
// return next number in sequence
//
void
TSobol::next ( TArray< double > & val, double max )
{
    val.set_size( _dim );
    
    // Find the position of the least-significant zero in count. */
    int  ell = 0;
    int  c   = _sequence_count;
    
    while ( true )
    {
        ++ell;
        
        if (( c % 2 ) == 1 ) c /= 2;
        else                 break;
    }

    // Check for exhaustion. */
    if ( ell > SOBOL_BIT_COUNT )
    {
        std::cout << "(TSobol) next : sequence exhausted" << std::endl;
        return;
    }// if

    for( uint i_dim = 0; i_dim < _dim; i_dim++ )
    {
        const int direction_i     = _v_direction[ell-1][i_dim];
        const int old_numerator_i = _last_numerator_vec[i_dim];
        const int new_numerator_i = old_numerator_i ^ direction_i;
        
        _last_numerator_vec[i_dim] = new_numerator_i;
        val[i_dim]                 = max * new_numerator_i * _last_denominator_inv;
    }

    _sequence_count++;
}

//
// return next number in sequence
// (special version for dim == 3)
//
void
TSobol::next ( TVector & v, double max )
{
    if ( _dim != 3 )
    {
        std::cout << "(TSobol) next : dimension of Sobol seq. not 3" << std::endl;
        return;
    }// if
    
    // Find the position of the least-significant zero in count. */
    int  ell = 0;
    int  c   = _sequence_count;
    
    while ( true )
    {
        ++ell;
        
        if (( c % 2 ) == 1 ) c /= 2;
        else                 break;
    }

    // Check for exhaustion. */
    if ( ell > SOBOL_BIT_COUNT )
    {
        std::cout << "(TSobol) next : sequence exhausted" << std::endl;
        return;
    }// if

    for( uint i_dim = 0; i_dim < 3; i_dim++ )
    {
        const int direction_i     = _v_direction[ell-1][i_dim];
        const int old_numerator_i = _last_numerator_vec[i_dim];
        const int new_numerator_i = old_numerator_i ^ direction_i;
        
        _last_numerator_vec[i_dim] = new_numerator_i;
        v[i_dim]                   = max * new_numerator_i * _last_denominator_inv;
    }

    _sequence_count++;
}

//
// initialise sobol sequence
//
void
TSobol::init ( uint dim )
{
    if (( dim < 1 ) || ( dim > 6 ))
    {
        std::cout << "(TSobol) init : invalid dimension; valid = 1 ... 6" << std::endl;
        return;
    }// if

    _dim = dim;
    
    // initialize direction table in dimension 0
    for( uint  k = 0; k < SOBOL_BIT_COUNT; k++ )
        _v_direction[k][0] = 1;

    // initialize in remaining dimensions.
    for( uint i_dim = 1; i_dim < _dim; i_dim++ )
    {
        const int poly_index = i_dim;
        const int degree_i   = degree_table[poly_index];
        int       includ[8];

        //
        // expand the polynomial bit pattern to separate
        // components of the logical array includ[].
        //
        
        int p_i = primitive_polynomials[poly_index];
        
        for( int k = degree_i-1; k >= 0; k-- )
        {
            includ[k]  = ((p_i % 2) == 1);
            p_i       /= 2;
        }// for

        // leading elements for dimension i come from v_init[][]
        for( int j = 0; j < degree_i; j++ )
            _v_direction[j][i_dim] = v_init[j][i_dim];

        //
        // calculate remaining elements for this dimension,
        // as explained in Bratley+Fox, section 2.
        //
        
        for( int j = degree_i; j < SOBOL_BIT_COUNT; j++ )
        {
            int newv = _v_direction[j-degree_i][i_dim];
            int ell  = 1;
            
            for( int k = 0; k < degree_i ; k++ )
            {
                ell *= 2;
                if ( includ[k] )
                    newv ^= ( ell * _v_direction[j-k-1][i_dim] );
            }// for
            
            _v_direction[j][i_dim] = newv;
        }// for
    }// for

    // multiply columns of v by appropriate power of 2
    int ell = 1;
    
    for( int j = SOBOL_BIT_COUNT-1-1; j >= 0; j-- )
    {
        ell *= 2;
        
        for( uint i_dim = 0; i_dim < _dim; i_dim++ )
            _v_direction[j][i_dim] *= ell;
    }// for

    // 1/(common denominator of the elements in v_direction) */
    _last_denominator_inv = 1.0 / (2.0 * ell);

    // final setup
    _sequence_count = 0;
    
    for ( uint i_dim = 0; i_dim < _dim; i_dim++ )
        _last_numerator_vec[i_dim] = 0;
}

}// namespace
