//
// Project : lib3D
// File    : TThinLensCamera.cc
// Purpose : Declaration of a class for central-projection
// Author  : Ronald Kriemann
//
// arch-tag: b02a77e1-fd21-499e-89dd-811781016b86
//

#include "defines.hh"
#include "TSobol.hh"
#include "TThinLensCamera.hh"

namespace RGL
{

//////////////////////////////////////////////
//
// constructor and destructor
//

TThinLensCamera::TThinLensCamera ()
{
    _focus_dist = -1;
    
    construct_by_dir( TVector( 0, 0, 1 ) );
    set_thin_lens( 0, -1, 1 );
}

TThinLensCamera::~TThinLensCamera ()
{
}

//
// define thin-lens of camera
//
void
TThinLensCamera::set_thin_lens ( real aperture, real focus_dist, uint n_samples )
{
    _focus_dist = focus_dist;
    
    // depth of field only with positive aperture
    if (( aperture = std::abs( aperture ) ) < EPSILON )
    {
        n_samples   = 1;
        _tl_radius  = 0.0;
        _focus_dist = -1;
    }// if
    else
    {
        _tl_radius = 0.5 * _film_dist / aperture;

        if ( verbose(5) )
            std::cout << "lens radius   = " << _tl_radius << std::endl;
    }// else

    //
    // update view plane with new parameters
    //
    
    update();

    //
    // setup random points on lens
    //
    
    TSobol          sobol( 2 );
    TArray< real >  sample( 2 );
    
    _tl_samples = max( n_samples, uint(1) );
    _tl_samples_x.set_size( _tl_samples );
    _tl_samples_y.set_size( _tl_samples );

    if ( _tl_samples == 1 )
    {
        _tl_samples_x[0] = 0.0;
        _tl_samples_y[0] = 0.0;
    }// if
    else
    {
        for ( uint i = 0; i < _tl_samples; i++ )
        {
            sobol.next( sample, 2.0 );
            
            _tl_samples_x[i] = _tl_radius * ( sample[0] - 1.0 );
            _tl_samples_y[i] = _tl_radius * ( sample[1] - 1.0 );
        }// for
    }// else
}
    
//
// update internal data
//
void
TThinLensCamera::update ()
{
    TCamera::update();
    
    //
    // if a focus distance was defined, this is used as the
    // distance for the view plane
    // everything has to be scaled w.r.t. the increased
    // distance to the view plane
    //

    if ( _focus_dist > 0.0 )
    {
        _width  = _film_width  * _focus_dist / _film_dist;
        _height = _film_height * _focus_dist / _film_dist;
        _front  = _film_dist / _focus_dist;

        if ( verbose(5) )
        {
            std::cout << "width  = " << _width   << std::endl;
            std::cout << "height = " << _height  << std::endl;
            std::cout << "front  = " << _front   << std::endl;
        }// if
    }// if

    //
    // set upper left corner, right and down vector of view plane
    //
    
    _ulc.set( 1.0, _origin, _focus_dist, _w ).add( -_width/2.0, _u, _height/2.0, _v );
    _vp_right.set( _width, _u );
    _vp_down.set( -_height, _v );
}

} // namespace

