//
// Project : 
// File    : ufo.cc
// Purpose : output management (cout,cerr redefinition)
// Author  : Ronald Kriemann
//
// arch-tag: 1a73571c-080b-41d0-b6fd-87ccd2e130d6
//

#include "ufo.hh"

namespace RGL
{
    
////////////////////////////////////////////////
//
// defines
//
////////////////////////////////////////////////

// colour values/attributes
#define ESC_BLACK  "30"
#define ESC_RED    "31"
#define ESC_GREEN  "32"
#define ESC_BROWN  "33"
#define ESC_BLUE   "34"
#define ESC_PURPLE "35"
#define ESC_CYAN   "36"
#define ESC_WHITE  "37"

#define ESC_NORMAL "0"
#define ESC_BOLD   "1"

#define ESC        "\033"

////////////////////////////////////////////////
//
// functions: forward decl.
//
////////////////////////////////////////////////

static bool check_terminal ();
static void set_messages ();

////////////////////////////////////////////////
//
// variables
//
////////////////////////////////////////////////

// output colours
TColourOut NORMAL;
TColourOut BOLD;
TColourOut RED;
TColourOut GREEN;
TColourOut BLUE;

// known colour terminals
static char colour_terms[][64] = { "xterm", "xterm-color", "xterm-256color",
                                   "colorxterm", "rxvt",
                                   "linux", "screen", "vt100",
                                   "\0" };                       // leave this at the end !!!

// use escape sequences to print messages
static bool use_esc_seq = check_terminal();

////////////////////////////////////////////////
//
// byte output
//
////////////////////////////////////////////////

// static variables
int TByteOutput::_verbosity = 0;

// object for easier usage
TByteOutput byte_out;

// byte -> string
TString
TByteOutput::operator () ( unsigned long count ) const
{
    const unsigned long  mag = 1000;
    TString              s;
    char                 buf[64];
    
    // less than 1kb
    if ( count < mag )
    {
        snprintf( buf, sizeof(buf), "%ld B", count );
        return s = buf;
    }// if

    // less than 1MB
    if ( count < (mag * mag))
    {
        long  kb = count / mag;
        long   b = count % mag;

        if ( _verbosity == 0 )
            snprintf( buf, sizeof(buf), "%ld kB", kb  );
        else if ( _verbosity == 1 )
            snprintf( buf, sizeof(buf), "%ld,%ld kB", kb, (b * 10) / 1000  );
        else
        {
            if      (b < 10)  snprintf( buf, sizeof(buf), "%ld.00%ld B", kb, b );
            else if (b < 100) snprintf( buf, sizeof(buf), "%ld.0%ld B", kb, b );
            else              snprintf( buf, sizeof(buf), "%ld.%ld B", kb, b );
        }// else
        
        return s = buf;
    }// if

    // less than 1GB
    if ( count < (mag * mag * mag))
    {
        count /= mag;
        
        long  mb = count / mag;
        long  kb = count % mag;

        if ( _verbosity == 0 )
            snprintf( buf, sizeof(buf), "%ld MB", mb );
        else if ( _verbosity == 1 )
            snprintf( buf, sizeof(buf), "%ld,%ld MB", mb, (kb * 10) / 1000 );
        else
        {
            if      (kb < 10)  snprintf( buf, sizeof(buf), "%ld.00%ld kB", mb, kb );
            else if (kb < 100) snprintf( buf, sizeof(buf), "%ld.0%ld kB", mb, kb );
            else               snprintf( buf, sizeof(buf), "%ld.%ld kB", mb, kb );
        }// else

        return s = buf;
    }// if

    // more than 1GB
    {
        count /= (mag * mag);
        
        long  gb = count / mag;
        long  mb = count % mag;

        if ( _verbosity == 0 )
            snprintf( buf, sizeof(buf), "%ld GB", gb );
        else if ( _verbosity == 1 )
            snprintf( buf, sizeof(buf), "%ld,%ld GB", gb, (mb * 10) / 1000 );
        else
        {
            if      (mb < 10)  snprintf( buf, sizeof(buf), "%ld.00%ld MB", gb, mb );
            else if (mb < 100) snprintf( buf, sizeof(buf), "%ld.0%ld MB", gb, mb );
            else               snprintf( buf, sizeof(buf), "%ld.%ld MB", gb, mb );
        }// else

        return s = buf;
    }// if
}

////////////////////////////////////////////////
//
// local functions (implementation)
//
////////////////////////////////////////////////

//
// check for terminal type
//
static bool
check_terminal ()
{
    char * term = getenv( "TERM" );

    use_esc_seq = false;

    if ( term == NULL )
        return use_esc_seq;

    //
    // check for known terminal types
    //

    uint  pos = 0;
    
    while ( ! colour_terms[pos][0] == '\0' )
    {
        if ( strcmp( term, colour_terms[pos] ) == 0 )
        {
            use_esc_seq = true;
            break;
        }// if

        ++pos;
    }// while

    set_messages();

    return use_esc_seq;
}

//
// set up messages
//
static void
set_messages ()
{
    if ( use_esc_seq )
    {
        NORMAL.set_colour_str( ESC "[" ESC_NORMAL "m" );
        BOLD.set_colour_str(   ESC "[" ESC_BOLD   "m" );
        RED.set_colour_str(    ESC "[" ESC_RED    "m" );
        GREEN.set_colour_str(  ESC "[" ESC_GREEN  "m" );
        BLUE.set_colour_str(   ESC "[" ESC_BLUE   "m" );
    }// if
    else
    {
        NORMAL.set_colour_str( "" );
        BOLD.set_colour_str(   "" );
        RED.set_colour_str(    "" );
        GREEN.set_colour_str(  "" );
        BLUE.set_colour_str(   "" );
    }// else
}

////////////////////////////////////////////////
//
// TColourOut (implementation)
//
////////////////////////////////////////////////

//
// colorise argument
//
TString
TColourOut::operator() ( const TString & str ) const
{
    return _col + str + NORMAL.colour_str();
}

//
// stream operator
//
std::ostream &
operator << ( std::ostream & os, const TColourOut & c )
{
    os << c.colour_str();
    return os;
}

} // namespace
