//
// Project : lib3D
// File    : TCSG.cc
// Purpose : class for handling constructive solid geometry
// Author  : Ronald Kriemann
//
// arch-tag: 45506648-1a6f-488f-b96e-9a2e5abfea98
//

#include "TCSG.hh"

namespace RGL
{

//////////////////////////////////////////////////
//
// constructor and destructor
//

TCSG::TCSG ( TGraphObject * left,
             TGraphObject * right,
             csg_type_t     type )
{
    _left  = left;
    _right = _right;
    _type  = type;
}

TCSG::~TCSG ()
{
    delete _left;
    delete _right;
}
    
//////////////////////////////////////////////////
//
// raytracing methods
//

//
// compute intersection with ray
// - return true if intersected, else false
//
bool
TCSG::intersect ( const TRay & ray, real & dist ) const
{
    bool  hit_left = false;
    
    if (( _left == NULL ) || ! _left->intersect( ray, dist ) )
    {
        //
        // if intersection or difference:
        //   no intersection with combinded object
        //

        if (( _type == CSG_INTERSECTION ) ||
            ( _type == CSG_DIFFERENCE ))
            return false;
    }// if
    else
        hit_left = true;

    if (( _right == NULL ) || ! _right->intersect( ray, dist ) )
    {
        //
        // if intersection: missing combined object
        // if union       : if missed left obj. missing comb. obj
        //

        if (( _type == CSG_INTERSECTION ) ||
            (( _type == CSG_UNION ) && ! hit_left ))
            return false;
    }// if

    return true;
}

//
// compute normal at point
//
bool
TCSG::calc_normal ( const TVector & point,
                    TVector       & normal ) const
{
    return false;
}

//////////////////////////////////////////////////
//
// misc. methods
//

//
// store bounding box defined by xy, xz and yz planes
// in given vectors
// - return false of bb is not defined (e.g. infinite object)
//
bool
TCSG::bounding_box ( TVector & min_bb, TVector & max_bb ) const
{
    TVector  min1_bb, max1_bb;

    //
    // compute bounding box of individual objects
    //
    
    if (( _left != NULL ) && ! _left->bounding_box( min_bb, max_bb ) )
        return false;

    if (( _right != NULL ) && ! _right->bounding_box( min1_bb, max1_bb ))
        return false;

    //
    // join both bounding boxes
    //

    if ( _type == CSG_UNION )
    {
        min_bb.set( min( min_bb.x(), min1_bb.x() ),
                    min( min_bb.y(), min1_bb.y() ),
                    min( min_bb.z(), min1_bb.z() ) );
        max_bb.set( max( max_bb.x(), max1_bb.x() ),
                    max( max_bb.y(), max1_bb.y() ),
                    max( max_bb.z(), max1_bb.z() ) );
    }// if
    else if ( _type == CSG_INTERSECTION )
    {
        min_bb.set( max( min_bb.x(), min1_bb.x() ),
                    max( min_bb.y(), min1_bb.y() ),
                    max( min_bb.z(), min1_bb.z() ) );
        max_bb.set( min( max_bb.x(), max1_bb.x() ),
                    min( max_bb.y(), max1_bb.y() ),
                    min( max_bb.z(), max1_bb.z() ) );
    }// if
    else if ( _type == CSG_DIFFERENCE )
    {
        // use bounding box of left object so
        // nothing changes
    }// if
    else
        std::cout << "(TCSG) bounding_box : unkown CSG type" << std::endl;

    return true;
}

//
// print object
//
void
TCSG::print () const
{
    if      ( _type == CSG_UNION        ) std::cout << "union (";
    else if ( _type == CSG_INTERSECTION ) std::cout << "intersection (";
    else if ( _type == CSG_DIFFERENCE   ) std::cout << "difference (";
    else                                  std::cout << "unknown CSG (";

    if ( _left != NULL ) _left->print();
    else                 std::cout << "none";

    std::cout << ", ";
    
    if ( _right != NULL ) _right->print();
    else                  std::cout << "none";

    std::cout << " )" << std::endl;
}
    
//////////////////////////////////////////////////
//
// virtual constructors
//

TGraphObject *
TCSG::clone () const
{
    return NULL;
}

TGraphObject *
TCSG::copy  () const
{
    return NULL;
}

}// namespace
