//
// Project : lib3D
// File    : TPlane.cc
// Purpose : class for a (infinite) plane
// Author  : Ronald Kriemann
//
// arch-tag: faca886c-eac0-4047-bda2-9d8520dc02ab
//

#include "TPlane.hh"

namespace RGL
{

/////////////////////////////////////////////////
//
// constructor and destructor
//

TPlane::TPlane ()
{
    _normal.set( 0, 1, 0 );
    _d = 0;
}

TPlane::TPlane ( const TVector & x0,
                 const TVector & x1,
                 const TVector & x2 )
{
    TVector  t1, t2;

    t1.set( 1.0, x1, -1.0, x0 );
    t2.set( 1.0, x2, -1.0, x0 );

    // calculate normal with cross-product
    _normal.set_cross( t1, t2 );

    if ( _normal.norm2() == 0.0 )
    {
        std::cout << "(TPlane) TPlane : given vectors are coplanar, plane not defined"
                  << std::endl;
        _normal = VEC_0;
        _d = 0.0;
    }// if
    else
    {
        _normal.normalise();
        _d = - dot( _normal, x0 );
    }// else
}

TPlane::TPlane ( const TVector & normal,
                 real            dist )
        : _normal( normal )
{
    if ( _normal.norm2() == 0.0 )
    {
        std::cout << "(TPlane) TPlane : given plane has undefined normal" << std::endl;
        _normal = VEC_0;
        _d = 0.0;
    }// if
    else
    {
        _normal.normalise();
        _d = dist;
    }// else
}

//////////////////////////////////////////////////
//
// raytracing methods
//

//
// compute intersection with ray
//
bool
TPlane::intersect ( const TRay & ray, real & dist ) const
{
    // eval plane-equation for intersection
    real tmp = dot( _normal, ray.direction() );

    // check if given line is parallel
    if (std::abs( tmp ) < EPSILON )
        return false;

    dist = - (dot( _normal, ray.origin() ) + _d) / tmp;

    return true;
}

//
// compute normal at point
//
bool
TPlane::calc_normal ( const TVector & point,
                      TVector       & normal_vec ) const
{
    // check if point is in plane
    real t = dot( _normal, point ) + _d;
    
    if (std::abs(t) < EPSILON)
    {
        normal_vec = _normal;
        return true;
    }// if
    else
        return false;
}

//////////////////////////////////////////////////
//
// misc. methods
//

//
// print object
//
void
TPlane::print () const
{
    std::cout << "plane: " << _normal << ", " << _d << std::endl;
}
    
//////////////////////////////////////////////////
//
// virtual constructors
//

TPlane *
TPlane::copy  () const
{
    TPlane * p = (TPlane*) TGraphObject::copy();

    p->_normal = _normal;
    p->_d      = _d;
    
    return p;
}

} // namespace

