//
// Project : lib3D
// File    : TTriangle.cc
// Purpose : class representing a triangle
// Author  : Ronald Kriemann
//
// arch-tag: 7e5dcc84-4aa2-42bb-aeef-5c7bb80c04d7
//

#include "TTriangle.hh"

namespace RGL
{

/////////////////////////////////////////////
//
// constructors
//

TTriangle::TTriangle ()
{
    _v0.set( 0, 0, 0 );
    _v1.set( 1, 0, 0 );
    _v2.set( 0, 1, 0 );
    
    comp_plane_eq();
}

TTriangle::TTriangle ( const TVector & v0,
                       const TVector & v1,
                       const TVector & v2 )
        : _v0(v0), _v1(v1), _v2(v2)
{
    comp_plane_eq();
}

//////////////////////////////////////////////////
//
// misc. methods
//

//
// store bounding box defined by xy, xz and yz planes
// in given vectors
// - return false of bb is not defined (e.g. infinite object)
//
bool
TTriangle::bounding_box ( TVector & min_bb, TVector & max_bb ) const
{
    min_bb[0] = min( _v0[0], _v1[0] ); min_bb[0] = min( min_bb[0], _v2[0] );
    min_bb[1] = min( _v0[1], _v1[1] ); min_bb[1] = min( min_bb[1], _v2[1] );
    min_bb[2] = min( _v0[2], _v1[2] ); min_bb[2] = min( min_bb[2], _v2[2] );

    max_bb[0] = max( _v0[0], _v1[0] ); max_bb[0] = max( max_bb[0], _v2[0] );
    max_bb[1] = max( _v0[1], _v1[1] ); max_bb[1] = max( max_bb[1], _v2[1] );
    max_bb[2] = max( _v0[2], _v1[2] ); max_bb[2] = max( max_bb[2], _v2[2] );

    return true;
}

//////////////////////////////////////////////////
//
// raytracing methods
//

//
// compute intersection with ray
// - return true if intersected, else false
// - transforms triangle into unit-triangle
//   with barycentric coordinates (u,v)
//
bool
TTriangle::intersect ( const TRay & ray, real & dist ) const
{
#if 0
    // eval plane-equation for intersection
    real tmp = dot( _normal, ray.direction() );

    // check if given line is parallel
    if (std::abs( tmp ) < EPSILON )
        return false;

    dist = - (dot( _normal, ray.origin() ) + _distance) / tmp;

    // std::cout << dist << std::endl;

    //TVector  ip( ray.origin() + dist * ray.direction() );

    // std::cout << ip << std::endl;
    
    //
    // check if point (v) is inside triangle (point in polygon)
    //
    
    real     abs_normal[3] = { std::abs( _normal.x() ),
                               std::abs( _normal.y() ),
                               std::abs( _normal.z() ) };
    int      dom;
    real     u, v;

    // determine dominant axis
    if ( abs_normal[0] > abs_normal[1] )
    {
        if ( abs_normal[0] > abs_normal[2] ) dom = 0;
        else                                 dom = 2;
    }// if
    else
    {
        if ( abs_normal[1] > abs_normal[2] ) dom = 1;
        else                                 dom = 2;
    }// else


    switch ( dom )
    {
        case 0 :  // x-axis is dominant
            u = ray.origin().y() + dist * ray.direction().y();
            v = ray.origin().z() + dist * ray.direction().z();

            if ((_v1.y() - u) * (_v1.z() - _v0.z()) < (_v1.z() - v) * (_v1.y() - _v0.y()))
                return false;
            
            if ((_v2.y() - u) * (_v2.z() - _v1.z()) < (_v2.z() - v) * (_v2.y() - _v1.y()))
                return false;
            
            if ((_v0.y() - u) * (_v0.z() - _v2.z()) < (_v0.z() - v) * (_v0.y() - _v2.y()))
                return false;
            
            break;

        case 1 : // y-axis is dominant
            u = ray.origin().x() + dist * ray.direction().x();
            v = ray.origin().z() + dist * ray.direction().z();

            if ((_v1.x() - u) * (_v1.z() - _v0.z()) < (_v1.z() - v) * (_v1.x() - _v0.x()))
                return false;
            
            if ((_v2.x() - u) * (_v2.z() - _v1.z()) < (_v2.z() - v) * (_v2.x() - _v1.x()))
                return false;
            
            if ((_v0.x() - u) * (_v0.z() - _v2.z()) < (_v0.z() - v) * (_v0.x() - _v2.x()))
                return false;
            
            break;

        case 2 : // z-axis is dominant
            u = ray.origin().x() + dist * ray.direction().x();
            v = ray.origin().y() + dist * ray.direction().y();

            if ((_v1.x() - u) * (_v1.y() - _v0.y()) < (_v1.y() - v) * (_v1.x() - _v0.x()))
                return false;
            
            if ((_v2.x() - u) * (_v2.y() - _v1.y()) < (_v2.y() - v) * (_v2.x() - _v1.x()))
                return false;
            
            if ((_v0.x() - u) * (_v0.y() - _v2.y()) < (_v0.y() - v) * (_v0.x() - _v2.x()))
                return false;
            
            break;

        default :
            // not possible to reach !!!
            std::cout << "(TTriangle) intersect : have reached impossible position"
                      << std::endl;
    }// switch
    
    return true;
    
#else
    TVector  e1, e2, p;
    real     det;

    e1.set( 1.0, _v1, -1.0, _v0 );
    e2.set( 1.0, _v2, -1.0, _v0 );
    p.set_cross( ray.direction(), e2 );
    det = dot( e1, p );
    
    // if determinant is zero => parallel ray
    if ( std::abs( det ) < EPSILON )
        return false;

    // compute coordinates
    real    inv_det = 1.0 / det;
    TVector t;
    real    u, v;

    t.set( 1.0, ray.origin(), -1.0, _v0 );
    u = dot( t, p ) * inv_det;
    if (( u < 0.0 ) || ( u > 1.0 ))
        return false;

    TVector  q;

    q.set_cross( t, e1 );
    
    v = dot( ray.direction(), q ) * inv_det;
    if (( v < 0.0 ) || ( v+u > 1.0 ))
        return false;

    // ray intersects with triangle, compute distance
    dist = dot( e2, q ) * inv_det;
    
    // compute first 
    return true;
#endif
}

//////////////////////////////////////////////////
//
// misc.
//

//
// compute normal from corners
//
void
TTriangle::comp_plane_eq ()
{
    TVector  e0, e1;

    e0.set( 1.0, _v1, -1.0, _v0 );
    e1.set( 1.0, _v2, -1.0, _v0 );

    _normal.set_cross( e0, e1 );

    real d = _normal.norm2();

    if ( d < EPSILON )
        std::cout << "(TTriangle) compute_normal : edges are colinear" << std::endl;
    else
        _normal.scale( 1.0 / d );

    _distance = - dot( _normal, _v0 );
}

}// namespace
