//
// Project : lib3D
// File    : light_nodes.hh
// Purpose : light related parse-nodes
// Author  : Ronald Kriemann
//
// arch-tag: 6d0eb05f-724e-474a-b40c-268d3b399f62
//

#include "TColour.hh"
#include "TAmbient.hh"
#include "TPointLight.hh"
#include "TSpotLight.hh"
#include "TParser.hh"

namespace RGL
{

///////////////////////////////////////////////////////
///////////////////////////////////////////////////////
//
// external functions
//
///////////////////////////////////////////////////////
///////////////////////////////////////////////////////

TColour & to_colour ( TParser::TNodeData * nd );

///////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////
//
// local nodedata-types
//
///////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////

//
// light data
//
class TNDLight : public TParser::TNodeData
{
protected:
    TLight  * _light;

public:
    TNDLight ( TLight * l ) : _light( l ) {}

    ~TNDLight () { delete _light; }
    
    TLight *       light ()       { return _light; }
    const TLight * light () const { return _light; }

    virtual uint type () const { return TYPE_LIGHT; }
    
    virtual bool is_type  ( uint t ) const
    {
        return ((t == TYPE_LIGHT) || TParser::TNodeData::is_type(t));
    }

    virtual TNodeData * copy () const
    {
        if ( _light != NULL ) return new TNDLight( _light->copy() );
        else                  return new TNDLight( NULL );
    }
};

///////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////
//
// local node-types
//
///////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////

//
// ambient light
//
class TAmbientNode : public TParser::TNode
{
protected:
    // colour of the light
    TParser::TNode  * _col;

public:
    TAmbientNode ( TParser::TNode * col ) : _col(col) {}

    ~TAmbientNode () { delete _col; }

    virtual uint type () const
    {
        if (( _col != NULL ) && _col->is_type( TYPE_COLOUR ) )
            return TYPE_LIGHT;
        else
            return TYPE_UNDEF;
    }
    
    TParser::TNodeData * eval () const
    {
        if ( _col == NULL )
            parse_error( "have NULL colour-node in eval", pos() );

        TParser::TNodeData * c = _col->eval();

        if ( c == NULL )
            parse_error( "got NULL in eval of first argument of ambientlight()",
                         _col->pos() );

        if ( ! c->is_type( TYPE_COLOUR ) )
            parse_error( "wrong type in first argument of ambientlight(), expected colour",
                         _col->pos() );

        TNDLight * light = new TNDLight( new TAmbient( to_colour( c ) ) );

        delete c;

        return light;
    }
};

//
// point light
//
class TPointLightNode : public TParser::TNode
{
protected:
    // position and colour of the light
    TParser::TNode  * _pos;
    TParser::TNode  * _col;

public:
    TPointLightNode ( TParser::TNode * pos,
                      TParser::TNode * col )
            : _pos(pos), _col(col)
    {}

    ~TPointLightNode ()
    {
        delete _pos;
        delete _col;
    }

    virtual uint type () const
    {
        if ((_pos != NULL) && ( _col != NULL ) && 
            _pos->is_type( TYPE_VECTOR ) &&
            _col->is_type( TYPE_COLOUR ) )
            return TYPE_LIGHT;
        else
            return TYPE_UNDEF;
    }
    
    TParser::TNodeData * eval () const
    {
        if ( _pos == NULL )
            parse_error( "first argument is NULL in pointlight()", pos() );

        if ( _col == NULL )
            parse_error( "second argument is NULL in pointlight()", pos() );

        TParser::TNodeData * p = _pos->eval();

        if ( p == NULL )
            parse_error( "got NULL in eval of first argument of pointlight()", _pos->pos() );

        if ( ! p->is_type( TYPE_VECTOR ) )
            parse_error( "wrong type in first argument of pointlight(), expected vector",
                         _pos->pos() );

        TParser::TNodeData * c = _col->eval();

        if ( c == NULL )
            parse_error( "got NULL in eval of second argument of pointlight()", _col->pos() );

        if ( ! c->is_type( TYPE_COLOUR ) )
            parse_error( "wrong type in second argument of pointlight(), expected colour",
                         _col->pos() );

        TNDLight * light = new TNDLight( new TPointLight( ((TNDVector*) p)->vector(),
                                                          to_colour( c ) ) );

        delete p;
        delete c;

        return light;
    }
};

//
// spot light
//
class TSpotLightNode : public TParser::TNode
{
protected:
    // position, direction, angle and colour of the light
    TParser::TNode  * _pos;
    TParser::TNode  * _dir;
    TParser::TNode  * _angle;
    TParser::TNode  * _col;

public:
    TSpotLightNode ( TParser::TNode * pos,
                     TParser::TNode * dir,
                     TParser::TNode * angle,
                     TParser::TNode * col )
            : _pos(pos), _dir(dir), _angle(angle), _col(col)
    {}

    ~TSpotLightNode ()
    {
        delete _pos;
        delete _dir;
        delete _angle;
        delete _col;
    }

    virtual uint type () const
    {
        if ((_pos != NULL) && (_dir != NULL) &&
            (_angle != NULL) && (_col != NULL) &&
            _pos->is_type( TYPE_VECTOR ) &&
            _dir->is_type( TYPE_VECTOR ) &&
            _angle->is_type( TYPE_NUMBER ) &&
            _col->is_type( TYPE_COLOUR ) )
            return TYPE_LIGHT;
        else
            return TYPE_UNDEF;
    }
    
    TParser::TNodeData * eval () const
    {
        if ( _pos == NULL )
            parse_error( "have NULL in first argument of spotlight()", pos() );

        if ( _dir == NULL )
            parse_error( "have NULL in second argument of spotlight()", pos() );

        if ( _angle == NULL )
            parse_error( "have NULL in third argument of spotlight()", pos() );

        if ( _col == NULL )
            parse_error( "have NULL in fourth argument of spotlight()", pos() );

        TParser::TNodeData * p = _pos->eval();

        if ( p == NULL )
            parse_error( "got NULL in eval of first argument of spotlight", _pos->pos() );

        if ( ! p->is_type( TYPE_VECTOR ) )
            parse_error( "wrong type in first argument of spotlight, expected vector",
                         _pos->pos() );

        TParser::TNodeData * d = _dir->eval();

        if ( d == NULL )
            parse_error( "got NULL in eval of second argument in spotlight", _dir->pos() );

        if ( ! d->is_type( TYPE_VECTOR ) )
            parse_error( "wrong type in first argument of spotlight, expected vector",
                         _dir->pos() );

        TParser::TNodeData * a = _angle->eval();

        if ( a == NULL )
            parse_error( "got NULL in eval of third argument in spotlight", _angle->pos() );

        if ( ! a->is_type( TYPE_NUMBER ) )
            parse_error( "wrong type in third argument of spotlight, expected number",
                         _angle->pos() );

        TParser::TNodeData * c = _col->eval();

        if ( c == NULL )
            parse_error( "got NULL in eval of fourth argument of spotlight", _col->pos() );

        if ( ! c->is_type( TYPE_COLOUR ) )
            parse_error( "wrong type in second argument of point-light, expected colour",
                         _col->pos() );

        TNDLight * light = new TNDLight( new TSpotLight( ((TNDVector*) p)->vector(),
                                                         ((TNDVector*) d)->vector(),
                                                         ((TNDNumber*) a)->number(),
                                                         to_colour( c ) ) );

        delete p;
        delete d;
        delete a;
        delete c;

        return light;
    }
};

//
// add the light to the world
//

class TAddLightNode : public TParser::TUnaryNode
{
protected:
    // world we add the object to
    TWorld * _world;
    
public:
    TAddLightNode ( TParser::TNode * son, TWorld * world )
            : TParser::TUnaryNode( son ), _world(world)
    {
        if ( _world == NULL )
            parse_warn( "can not add to NULL world", pos() );
    }

    virtual uint type () const
    {
        return (_son != NULL ? _son->type() : TYPE_UNDEF);
    }
    
    TParser::TNodeData * eval () const
    {
        if ( _son == NULL )
            return NULL;

        TParser::TNodeData  * dat = NULL;

        dat = _son->eval();

        if ( dat == NULL )
        {
            parse_error( "got NULL data in eval", _son->pos() );
            return NULL;
        }// if
        
        if ( dat->is_type( TYPE_LIGHT ) )
        {
            if ( _world == NULL )
                return dat;

            // add a copy, not the original object
            _world->add_light( ((TNDLight*) dat)->light()->copy() );

            return dat;
        }// if
        else
        {
            parse_error( "can only add lights to world", _son->pos() );
            delete dat;
            return NULL;
        }// else
        
        return NULL;
    }
};

///////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////
//
// return local node types to outer world
//
///////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////

TParser::TNode * new_ambient_light ( TParser::TNode * n )
{
    return new TAmbientNode( n );
}

TParser::TNode * new_point_light ( TParser::TNode * n0,
                                   TParser::TNode * n1 )
{
    return new TPointLightNode( n0, n1 );
}

TParser::TNode * new_spot_light ( TParser::TNode * n0,
                                  TParser::TNode * n1,
                                  TParser::TNode * n2,
                                  TParser::TNode * n3 )
{
    return new TSpotLightNode( n0, n1, n2, n3 );
}

TParser::TNode * new_addlight ( TParser::TNode * n,
                                TWorld         * world )
{
    return new TAddLightNode( n, world );
}

}// namespace
