//
// Project : lib3D
// File    : func_nodes.hh
// Purpose : functional parse-nodes
// Author  : Ronald Kriemann
//
// arch-tag: 5aba0cb2-2632-4a62-87ed-e5b554d3fa5a
//

#include "TSphere.hh"
#include "TPlane.hh"
#include "TTriangle.hh"
#include "TCube.hh"
#include "TCylinder.hh"
#include "TLight.hh"
#include "TParser.hh"

namespace RGL
{

///////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////
//
// object types
//
///////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////

//
// node for a number
//

class TNumberNode : public TParser::TNode
{
protected:
    real  _number;

public:
    TNumberNode ( real f ) : _number(f) {}

    virtual uint type () const { return TYPE_NUMBER; }
        
    TParser::TNodeData * eval () const
    {
        return new TNDNumber( _number );
    }
};

//
// node for a boolean
//

class TBoolNode : public TParser::TNode
{
protected:
    bool  _boolean;

public:
    TBoolNode ( bool b ) : _boolean(b) {}

    virtual uint type () const { return TYPE_BOOL; }
        
    TParser::TNodeData * eval () const
    {
        return new TNDBool( _boolean );
    }
};

//
// node for a string
//

class TStringNode : public TParser::TNode
{
protected:
    TString  _string;

public:
    TStringNode ( const TString & s ) : _string(s) {}

    virtual uint type () const { return TYPE_STRING; }
        
    TParser::TNodeData * eval () const
    {
        return new TNDString( _string );
    }
};

//
// node for a vector
//

class TVectorNode : public TParser::TNode
{
protected:
    TParser::TNode * _vector[3];

public:
    TVectorNode ( TParser::TNode * n0,
                  TParser::TNode * n1,
                  TParser::TNode * n2 )
    {
        _vector[0] = n0;
        _vector[1] = n1;
        _vector[2] = n2;
    }

    ~TVectorNode () { for ( uint i = 0; i < 3; i++ ) delete _vector[i]; }
    
    virtual uint type () const { return TYPE_VECTOR; }
    
    TParser::TNodeData * eval () const
    {
        for ( uint i = 0; i < 3; i++ )
        {
            if ( _vector[i] == NULL )
            {
                parse_warn( "NULL node in vector", pos() );
                return NULL;
            }// if
        }// for
        
        TVector               v;
        TParser::TNodeData  * d0 = _vector[0]->eval();
        TParser::TNodeData  * d1 = _vector[1]->eval();
        TParser::TNodeData  * d2 = _vector[2]->eval();

        if ( d0->is_type( TYPE_NUMBER ) )
            v[0] = ((TNDNumber *) d0)->number();
        else
            parse_warn( "entry 0 is not a number", _vector[0]->pos() );
        
        if ( d1->is_type( TYPE_NUMBER ) )
            v[1] = ((TNDNumber *) d1)->number();
        else
            parse_warn( "entry 1 is not a number", _vector[1]->pos() );
        
        if ( d2->is_type( TYPE_NUMBER ) )
            v[2] = ((TNDNumber *) d2)->number();
        else
            parse_warn( "entry 2 is not a number", _vector[2]->pos() );
        
        delete d2;
        delete d1;
        delete d0;
        
        return new TNDVector( v );
    }
};

//
// node for a sphere
//

class TSphereNode : public TParser::TNode
{
protected:
    TParser::TNode  * _node0, * _node1;
    
public:
    TSphereNode ( TParser::TNode * n0,
                  TParser::TNode * n1 )
            : _node0(n0), _node1(n1)
    {}

    ~TSphereNode () { delete _node0; delete _node1; }
    
    virtual uint type () const { return TYPE_SPHERE; }
    
    TParser::TNodeData * eval () const
    {
        TVector  center;
        real     radius = 0.0;
        bool     have_radius = false;

        // check ordering and types of nodes
        if ( (_node0 != NULL) && (_node1 != NULL) )
        {
            if ( ! ( _node0->is_type( TYPE_VECTOR ) &&
                     _node1->is_type( TYPE_NUMBER ) ))
                parse_error( "wrong arguments (expected vector,number)", _node0->pos() );
        }// if

        //
        // eval nodes
        //
        
        if ( _node0 != NULL )
        {
            TParser::TNodeData * data0 = _node0->eval();

            if (data0 != NULL)
            {
                if ( data0->is_type( TYPE_NUMBER ) )
                {
                    radius      = ((TNDNumber*) data0)->number();
                    have_radius = true;
                }// if
                else if ( data0->is_type( TYPE_VECTOR ) )
                {
                    center = ((TNDVector*) data0)->vector();
                }// if
                else
                    parse_error( "unsupported expression", _node0->pos() );
            }// if
        }// if

        if ( _node1 != NULL )
        {
            TParser::TNodeData * data1 = _node1->eval();

            if (data1 != NULL)
            {
                if ( data1->is_type( TYPE_NUMBER ) )
                {
                    if ( have_radius )
                        parse_error( "wrong argument (2x number)", _node1->pos() );
                    else
                        radius = ((TNDNumber*) data1)->number();
                }// if
                else if ( data1->is_type( TYPE_VECTOR ) )
                    parse_error( "got vector, expected number", _node1->pos() );
                else
                    parse_error( "unsupported expression", _node1->pos() );
            }// if
        }// if

        return new TNDGraphObject( new TSphere( center, radius ) );
    }
};

//
// node for a plane
//

class TPlaneNode : public TParser::TNode
{
protected:
    TParser::TNode  * _n0, * _n1, * _n2;

    // type of plane-definition
    // 0: 3 points, 1: normal + distance
    uint              _type;
    
public:
    TPlaneNode ( TParser::TNode * n0,
                 TParser::TNode * n1,
                 TParser::TNode * n2 )
            : _n0(n0), _n1(n1), _n2(n2), _type(0)
    {}

    TPlaneNode ( TParser::TNode * n0,
                 TParser::TNode * n1 )
            : _n0(n0), _n1(n1), _n2(NULL), _type(1)
    {}

    ~TPlaneNode () { delete _n0; delete _n1; delete _n2; }
    
    virtual uint type () const
    {
        if (( _type == 0 ) &&
            (_n0 != NULL) && (_n1 != NULL) && (_n2 != NULL) &&
            _n0->is_type( TYPE_VECTOR ) &&
            _n1->is_type( TYPE_VECTOR ) &&
            _n2->is_type( TYPE_VECTOR ))
            return TYPE_GRAPH_OBJECT;
        else if (( _type == 1 ) &&
                 (_n0 != NULL) && (_n1 != NULL) &&
                 _n0->is_type( TYPE_VECTOR ) &&
                 _n1->is_type( TYPE_NUMBER ))
            return TYPE_GRAPH_OBJECT;
        else
            return TYPE_UNDEF;
    }
    
    TParser::TNodeData * eval () const
    {
        if ( _type == 0 )
        {
            if ( _n0 == NULL )
                parse_error( "first argument is NULL in plane()", pos() );

            if ( _n1 == NULL )
                parse_error( "second argument is NULL in plane()", pos() );

            if ( _n2 == NULL )
                parse_error( "third argument is NULL in plane()", pos() );

            //
            // eval nodes
            //
            
            TParser::TNodeData  * dat0, * dat1, * dat2;

            if (( dat0 = _n0->eval() ) == NULL )
                parse_error( "got NULL in eval of first argument of plane()",
                             _n0->pos() );
            
            if ( ! dat0->is_type( TYPE_VECTOR ) )
                parse_error( "wrong value in first argument of plane(), expected vector",
                             _n0->pos() );

            if (( dat1 = _n1->eval() ) == NULL )
                parse_error( "got NULL in eval of second argument of plane()",
                             _n1->pos() );
            
            if ( ! dat1->is_type( TYPE_VECTOR ) )
                parse_error( "wrong value in second argument of plane(), expected vector",
                             _n1->pos() );

            if (( dat2 = _n2->eval() ) == NULL )
                parse_error( "got NULL in eval of third argument of plane()",
                             _n2->pos() );
            
            if ( ! dat2->is_type( TYPE_VECTOR ) )
                parse_error( "wrong value in third argument of plane(), expected vector",
                             _n2->pos() );

            //
            // ok, three vectors, ready to build plane
            //

            TNDGraphObject  * res;

            res = new TNDGraphObject( new TPlane( ((TNDVector*) dat0)->vector(),
                                                  ((TNDVector*) dat1)->vector(),
                                                  ((TNDVector*) dat2)->vector() ) );

            delete dat0;
            delete dat1;
            delete dat2;

            return res;
        }// if
        else if ( _type == 1 )
        {
            if ( _n0 == NULL )
                parse_error( "first argument is NULL in plane()", pos() );

            if ( _n1 == NULL )
                parse_error( "second argument is NULL in plane()", pos() );

            //
            // eval nodes
            //
            
            TParser::TNodeData  * dat0, * dat1;

            if (( dat0 = _n0->eval() ) == NULL )
                parse_error( "got NULL in eval of first argument of plane()",
                             _n0->pos() );
            
            if ( ! dat0->is_type( TYPE_VECTOR ) )
                parse_error( "wrong value in first argument of plane(), expected vector",
                             _n0->pos() );

            if (( dat1 = _n1->eval() ) == NULL )
                parse_error( "got NULL in eval of second argument of plane()",
                             _n1->pos() );
            
            if ( ! dat1->is_type( TYPE_NUMBER ) )
                parse_error( "wrong value in second argument of plane(), expected number",
                             _n1->pos() );

            //
            // ok, a vector and a number, ready to build plane
            //

            TNDGraphObject  * res;

            res = new TNDGraphObject( new TPlane( ((TNDVector*) dat0)->vector(),
                                                  ((TNDNumber*) dat1)->number() ) );

            delete dat0;
            delete dat1;

            return res;
        }// if
        else
            parse_error( "wrong plane-type", pos() );
        
        return NULL;
    }
};

//
// node for a triangle
//

class TTriangleNode : public TParser::TNode
{
protected:
    TParser::TNode  * _n0, * _n1, * _n2;

public:
    TTriangleNode ( TParser::TNode * n0,
                    TParser::TNode * n1,
                    TParser::TNode * n2 )
            : _n0(n0), _n1(n1), _n2(n2)
    {}

    ~TTriangleNode () { delete _n0; delete _n1; delete _n2; }
    
    virtual uint type () const
    {
        if ((_n0 != NULL) && (_n1 != NULL) && (_n2 != NULL) &&
            _n0->is_type( TYPE_VECTOR ) &&
            _n1->is_type( TYPE_VECTOR ) &&
            _n2->is_type( TYPE_VECTOR ))
            return TYPE_GRAPH_OBJECT;
        else
            return TYPE_UNDEF;
    }
    
    TParser::TNodeData * eval () const
    {
        if ( _n0 == NULL )
            parse_error( "first argument is NULL in triangle()", pos() );

        if ( _n1 == NULL )
            parse_error( "second argument is NULL in triangle()", pos() );

        if ( _n2 == NULL )
            parse_error( "third argument is NULL in triangle()", pos() );

        //
        // eval nodes
        //
            
        TParser::TNodeData  * dat0, * dat1, * dat2;
        
        if (( dat0 = _n0->eval() ) == NULL )
            parse_error( "got NULL in eval of first argument of triangle()",
                         _n0->pos() );
        
        if ( ! dat0->is_type( TYPE_VECTOR ) )
            parse_error( "wrong value in first argument of triangle(), expected vector",
                         _n0->pos() );
        
        if (( dat1 = _n1->eval() ) == NULL )
            parse_error( "got NULL in eval of second argument of triangle()",
                         _n1->pos() );
        
        if ( ! dat1->is_type( TYPE_VECTOR ) )
            parse_error( "wrong value in second argument of triangle(), expected vector",
                         _n1->pos() );
        
        if (( dat2 = _n2->eval() ) == NULL )
            parse_error( "got NULL in eval of third argument of triangle()",
                         _n2->pos() );
        
        if ( ! dat2->is_type( TYPE_VECTOR ) )
            parse_error( "wrong value in third argument of triangle(), expected vector",
                         _n2->pos() );
        
        //
        // ok, three vectors, ready to build triangle
        //

        TNDGraphObject  * res;

        res = new TNDGraphObject( new TTriangle( ((TNDVector*) dat0)->vector(),
                                                 ((TNDVector*) dat1)->vector(),
                                                 ((TNDVector*) dat2)->vector() ) );

        delete dat0;
        delete dat1;
        delete dat2;
        
        return res;
    }
};

//
// node for a cube
//

class TCubeNode : public TParser::TNode
{
public:
    TCubeNode () {}

    ~TCubeNode () {}
    
    virtual uint type () const
    {
        return TYPE_GRAPH_OBJECT;
    }
    
    TParser::TNodeData * eval () const
    {
        return new TNDGraphObject( new TCube() );
    }
};

//
// node for a cylinder
//

class TCylinderNode : public TParser::TNode
{
protected:
    TParser::TNode  * _closed;
    
public:
    TCylinderNode ( TParser::TNode * closed )
            : _closed(closed)
    {}

    ~TCylinderNode () { delete _closed; }
    
    virtual uint type () const
    {
        if ((_closed != NULL) && _closed->is_type( TYPE_BOOL ))
            return TYPE_GRAPH_OBJECT;
        else
            return TYPE_UNDEF;
    }
    
    TParser::TNodeData * eval () const
    {
        if ( _closed == NULL )
            parse_error( "have NULL as first argument of cylinder()", pos() );

        TParser::TNodeData * c = _closed->eval();

        if ( c == NULL )
            parse_error( "got NULL in eval of first argument of cylinder()", pos() );

        if ( ! c->is_type( TYPE_BOOL ) )
            parse_error( "wrong argument in first argument of cylinder(), expected boolean",
                         pos() );

        TNDGraphObject * res;

        res = new TNDGraphObject( new TCylinder( ((TNDBool*) c)->boolean() ) );

        delete c;

        return res;
    }
};

//
// adds given object to world
//
class TAddWorldNode : public TParser::TUnaryNode
{
protected:
    // world we add the object to
    TWorld * _world;
    
public:
    TAddWorldNode ( TParser::TNode * son, TWorld * world )
            : TParser::TUnaryNode( son ), _world(world)
    {
        if ( _world == NULL )
            parse_warn( "can not add to NULL world", pos() );
    }

    virtual uint type () const
    {
        return (_son != NULL ? _son->type() : TYPE_UNDEF);
    }
    
    TParser::TNodeData * eval () const
    {
        if ( _son == NULL )
            return NULL;

        TParser::TNodeData  * dat = NULL;

        dat = _son->eval();

        if ( dat == NULL )
        {
            parse_error( "got NULL data in eval", _son->pos() );
            return NULL;
        }// if
        
        if ( dat->is_type( TYPE_GRAPH_OBJECT ) )
        {
            if ( _world == NULL )
                return dat;

            // add a copy, not the original object
            _world->add( ((TNDGraphObject*) dat)->graph_object()->copy() );

            return dat;
        }// if
        else
        {
            parse_error( "can only add graphical objects to world", _son->pos() );
            delete dat;
            return NULL;
        }// else
        
        return NULL;
    }
};

//
// apply transformations and textures to graphical objects
//

// external apply functions
void apply_trans  ( TGraphObject * obj, TParser::TNodeData * nd );
void apply_colour ( TGraphObject * obj, TParser::TNodeData * nd );

class TApplyNode : public TParser::TBinaryNode
{
public:
    TApplyNode ( TParser::TNode * son0, TParser::TNode * son1 )
            : TParser::TBinaryNode( son0, son1 )
    {}

    virtual uint type () const
    {
        if (( _son0 != NULL ) && (_son1 != NULL) &&
            _son0->is_type( TYPE_GRAPH_OBJECT ) &&
            _son1->is_type( TYPE_TRANSFORM ))
            return TYPE_GRAPH_OBJECT;
        else
            return TYPE_UNDEF;
    }
    
    TParser::TNodeData * eval () const
    {
        if ( _son0 == NULL )
        {
            parse_warn( "first argument NULL in eval of apply", pos() );
            return NULL;
        }// if
        
        if ( _son1 == NULL )
        {
            parse_warn( "second argument NULL in eval of apply", pos() );
            return NULL;
        }// if
        
        //
        // check if first node is a variable
        //
        
        TParser::TVarNode * var = dynamic_cast< TParser::TVarNode * >( _son0 );
        
        if ( var != NULL )
        {
            if ( ! var->is_type( TYPE_GRAPH_OBJECT ) )
            {
                parse_error( "can not apply to non-graphical objects", var->pos() );
                return NULL;
            }// if

            if ( var->variable() == NULL ) 
            {
                parse_error( "no variable associated with node", var->pos() );
                return NULL;
            }// if

            if ( var->variable()->data() == NULL ) 
            {
                parse_error( "variable not initialised", var->pos() );
                return NULL;
            }// if

            TParser::TNodeData * dat = _son1->eval();

            if ( dat == NULL )
            {
                parse_error( "got NULL in eval of apply", _son1->pos() );
                return NULL;
            }// if

            if ( dat->is_type( TYPE_TRANSFORM ) )
            {
                apply_trans( ((TNDGraphObject*) var->variable()->data())->graph_object(),
                             dat );
            }// if
            else if ( dat->is_type( TYPE_COLOUR )    ||
                      dat->is_type( TYPE_COL_COEFF ) ||
                      dat->is_type( TYPE_TEXTURE ) )
            {
                apply_colour( ((TNDGraphObject*) var->variable()->data())->graph_object(),
                              dat );
            }// if
            else
            {
                parse_error( "expected transformation or texture in apply",
                             _son1->pos() );
                delete dat;
                return NULL;
            }// else

            delete dat;
            
            // var->variable()->print( var->pos() );
            return var->eval();
        }// if
        else
        {
            parse_error( "can't apply to constant", _son0->pos() );
            return NULL;
        }// else
    }
};

//
// node for camera definition
//
class TCameraNode : public TParser::TNode
{
protected:
    // parser to set camera-data
    TParser         * _parser;

    TParser::TNode  * _orig, * _dir, * _up;

public:
    TCameraNode ( TParser        * parser,
                  TParser::TNode * orig,
                  TParser::TNode * dir,
                  TParser::TNode * up )
            : _parser(parser), _orig(orig), _dir(dir), _up(up)
    {
        if ( _parser == NULL )
            parse_error( "(TCameraNode) got NULL parser", pos() );
    }

    ~TCameraNode () { delete _orig; delete _dir; delete _up; }

    virtual uint type () const
    {
        return TYPE_UNDEF;
    }
    
    TParser::TNodeData * eval () const
    {
        if ( _orig == NULL )
            parse_error( "have NULL as first argument of camera()", pos() );

        if ( _dir == NULL )
            parse_error( "have NULL as second argument of camera()", pos() );

        TParser::TNodeData  * orig, * dir;

        if ((orig = _orig->eval()) == NULL)
            parse_error( "got NULL in eval of first argument of camera()", _orig->pos() );

        if ( ! orig->is_type( TYPE_VECTOR ) )
            parse_error( "wrong return-type in eval of first argument of camera()",
                         _orig->pos() );

        if ((dir = _dir->eval()) == NULL)
            parse_error( "got NULL in eval of second argument of camera()", _dir->pos() );

        if ( ! dir->is_type( TYPE_VECTOR ) )
            parse_error( "wrong return-type in eval of second argument of camera()",
                         _orig->pos() );

        _parser->set_camera( ((TNDVector *) orig)->vector(),
                             ((TNDVector *) dir)->vector() );

        if ( _up != NULL )
        {
            TParser::TNodeData  * up;

            if ((up = _up->eval()) == NULL)
                parse_error( "got NULL in eval of third argument of camera()", _up->pos() );

            if ( ! up->is_type( TYPE_VECTOR ) )
                parse_error( "wrong return-type in eval of third argument of camera()",
                             _up->pos() );

            _parser->set_camera( ((TNDVector *) orig)->vector(),
                                 ((TNDVector *) dir)->vector() );

            delete up;
        }// if
        
        delete orig;
        delete dir;

        return NULL;
    }
};

///////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////
//
// return local node types to outer world
//
///////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////

TParser::TNode * new_number ( real f )
{
    return new TNumberNode( f );
}

TParser::TNode * new_bool ( bool b )
{
    return new TBoolNode( b );
}

TParser::TNode * new_string ( const TString & buf )
{
    TString tmp( buf );
    
    // remove " from string
    if ( tmp[0] == '\"' )
        tmp = tmp.substr( 1, tmp.length()-1 );

    if ( tmp[ tmp.length()-1] == '\"' )
        tmp = tmp.substr( 0, tmp.length()-1 );
    
    return new TStringNode( tmp );
}

TParser::TNode * new_vector ( TParser::TNode * n0,
                              TParser::TNode * n1,
                              TParser::TNode * n2 )
{
    return new TVectorNode( n0, n1, n2 );
}

TParser::TNode * new_sphere ( TParser::TNode * n0,
                              TParser::TNode * n1 )
{
    return new TSphereNode( n0, n1 );
}

TParser::TNode * new_plane ( TParser::TNode * n0,
                             TParser::TNode * n1,
                             TParser::TNode * n2 )
{
    return new TPlaneNode( n0, n1, n2 );
}

TParser::TNode * new_plane ( TParser::TNode * n0,
                             TParser::TNode * n1 )
{
    return new TPlaneNode( n0, n1 );
}

TParser::TNode * new_triangle ( TParser::TNode * n0,
                                TParser::TNode * n1,
                                TParser::TNode * n2 )
{
    return new TTriangleNode( n0, n1, n2 );
}

TParser::TNode * new_cube ()
{
    return new TCubeNode();
}

TParser::TNode * new_cylinder ( TParser::TNode * n )
{
    return new TCylinderNode( n );
}

TParser::TNode * new_addworld ( TParser::TNode * n,
                                TWorld         * world )
{
    return new TAddWorldNode( n, world );
}

TParser::TNode * new_apply ( TParser::TNode * n0,
                             TParser::TNode * n1 )
{
    return new TApplyNode( n0, n1 );
}

TParser::TNode * new_camera ( TParser        * parser,
                              TParser::TNode * orig,
                              TParser::TNode * dir,
                              TParser::TNode * up )
{
    return new TCameraNode( parser, orig, dir, up );
}

}// namespace
