//
// Project : lib3D
// File    : tex_nodes.hh
// Purpose : texture related parse-nodes
// Author  : Ronald Kriemann
//
// arch-tag: 6c10a65e-e320-4c4f-9a98-07d65dbae218
//

#include "TColour.hh"
#include "TParser.hh"

namespace RGL
{

// texture values
#define TEX_COLOUR      0
#define TEX_AMBIENT     1
#define TEX_DIFFUSE     2
#define TEX_SPECULAR    3
#define TEX_TRANSMISS   4
#define TEX_IOR         5

#define MAX_TEX_VAL     6

///////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////
//
// local nodedata-types
//
///////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////

//
// colour-values
//
class TNDColour : public TParser::TNodeData
{
protected:
    // the colour value
    TColour  _colour;

public:
    TNDColour () {}

    virtual ~TNDColour () {}
    
    TColour &       colour ()       { return _colour; }
    const TColour & colour () const { return _colour; }

    virtual uint type () const { return TYPE_COLOUR; }
    virtual bool is_type  ( uint t ) const
    { return ((t == TYPE_COLOUR) || TParser::TNodeData::is_type(t)); }
    
    virtual TNodeData * copy () const
    {
        TNDColour  * c = new TNDColour;

        c->colour() = _colour;
        return c;
    }
};

//
// colour coefficient
//
class TNDColCoeff : public TParser::TNodeData
{
protected:
    uint  _tex_type;
    real  _coeff1;
    real  _coeff2;
    
public:
    TNDColCoeff ( uint tex_type, real c1, real c2 = 0.0 )
            : _tex_type(tex_type), _coeff1(c1), _coeff2(c2) {}

    virtual ~TNDColCoeff () {}

    uint tex_type () const { return _tex_type; }
    real coeff1   () const { return _coeff1; }
    real coeff2   () const { return _coeff2; }

    virtual uint type () const { return TYPE_COL_COEFF; }
    virtual bool is_type  ( uint t ) const
    { return ((t == TYPE_COL_COEFF) || TParser::TNodeData::is_type(t)); }
    
    virtual TNodeData * copy () const
    { return new TNDColCoeff( _tex_type, _coeff1, _coeff2 ); }
};

//
// complete texture data
//
class TNDTexture : public TParser::TNodeData
{
protected:
    TTexture * _tex;

    // true if already set
    bool       _tex_vals[ MAX_TEX_VAL ];
    
public:
    TNDTexture ()
    {
        _tex = new TTexture;
        _tex->add_ref();
        
        for ( uint i = 0; i < MAX_TEX_VAL; i++ )
            _tex_vals[i] = false;
    }
    
    TNDTexture ( TTexture * t )
    {
        if ( t != NULL )
            t->add_ref();
        _tex = t;
    }

    virtual ~TNDTexture () { if ( _tex != NULL ) _tex->free_ref(); }

    TTexture *       texture ()       { return _tex; }
    const TTexture * texture () const { return _tex; }

    bool has_tex_val ( uint i ) const { return _tex_vals[i]; }
    void set_tex_val ( uint i, bool b ) { _tex_vals[i] = b; }

    virtual uint type () const { return TYPE_TEXTURE; }
    virtual bool is_type  ( uint t ) const
    { return ((t == TYPE_TEXTURE) || TParser::TNodeData::is_type(t)); }
    
    virtual TNodeData * copy () const
    {
        TNDTexture  * tex = new TNDTexture( _tex );

        for ( uint i = 0; i < MAX_TEX_VAL; i++ )
            tex->set_tex_val( i, has_tex_val(i) );

        return tex;
    }
};

///////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////
//
// local node-types
//
///////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////

//
// represents a colour-entry (RGB or HSV)
//

class TColourNode : public TParser::TNode
{
protected:
    // name of colour
    TParser::TNode  * _name;
    
    // colour components
    TParser::TNode  * _c0, * _c1, * _c2;

    // type of colour (default is RGB)
    bool              _is_hsv;
    bool              _is_named;

public:
    TColourNode ( TParser::TNode * c0,
                  TParser::TNode * c1,
                  TParser::TNode * c2,
                  bool             hsv )
            : _c0(c0), _c1(c1), _c2(c2), _is_hsv(hsv)
    {}

    TColourNode ( TParser::TNode * n )
            : _name(n), _is_hsv(false), _is_named(true)
    {}

    virtual uint type () const
    {
        if (( _c0 != NULL ) && ( _c1 != NULL ) && ( _c2 != NULL ) &&
            _c0->is_type( TYPE_NUMBER ) &&
            _c1->is_type( TYPE_NUMBER ) &&
            _c2->is_type( TYPE_NUMBER ))
            return TYPE_COLOUR;
        else if (( _name != NULL ) && _name->is_type( TYPE_STRING ) && _is_named )
            return TYPE_COLOUR;
        else
            return TYPE_UNDEF;
    }

    TParser::TNodeData * eval () const
    {
        if ( _is_named )
        {
            if ( _name == NULL )
            {
                parse_error( "argument 1 is NULL", pos() );
                return NULL;
            }// if

            TParser::TNodeData  * dat;

            if ((dat = _name->eval()) == NULL)
            {
                parse_error( "got NULL in eval of argument 1", _name->pos() );
                return NULL;
            }// if

            if ( ! dat->is_type( TYPE_STRING ) )
            {
                parse_error( "wrong type of argument 1, expected string", _name->pos() );
                delete dat;
                return NULL;
            }// if

            TNDColour * col = new TNDColour;
        
            col->colour().set_named( ((TNDString*) dat)->string().c_str() );

            delete dat;

            return col;
        }// if
        else
        {
            if ( _c0 == NULL )
            {
                parse_error( "argument 1 is NULL", pos() );
                return NULL;
            }// if

            if ( _c1 == NULL )
            {
                parse_error( "argument 2 is NULL", pos() );
                return NULL;
            }// if

            if ( _c2 == NULL )
            {
                parse_error( "argument 3 is NULL", pos() );
                return NULL;
            }// if

            TParser::TNodeData  * d0, * d1, * d2;

            if ((d0 = _c0->eval()) == NULL)
            {
                parse_error( "got NULL in eval of argument 1", _c0->pos() );
                return NULL;
            }// if

            if ( ! d0->is_type( TYPE_NUMBER ) )
            {
                parse_error( "wrong type of argument 1, expected number", _c0->pos() );
                delete d0;
                return NULL;
            }// if
        
            if ((d1 = _c1->eval()) == NULL)
            {
                parse_error( "got NULL in eval of argument 2", _c1->pos() );
                delete d0;
                return NULL;
            }// if

            if ( ! d1->is_type( TYPE_NUMBER ) )
            {
                parse_error( "wrong type of argument 2, expected number", _c1->pos() );
                delete d0;
                delete d1;
                return NULL;
            }// if
        
            if ((d2 = _c2->eval()) == NULL)
            {
                parse_error( "got NULL in eval of argument 3", _c2->pos() );
                delete d0;
                delete d1;
                return NULL;
            }// if

            if ( ! d2->is_type( TYPE_NUMBER ) )
            {
                parse_error( "wrong type of argument 3, expected number", _c2->pos() );
                delete d0;
                delete d1;
                delete d2;
                return NULL;
            }// if
        
            //
            // everything is as it should be => compose the colour value
            //

            TNDColour * col = new TNDColour;
            real       v0 = ((TNDNumber*) d0)->number();
            real       v1 = ((TNDNumber*) d1)->number();
            real       v2 = ((TNDNumber*) d2)->number();
        
            if ( _is_hsv ) col->colour().set_hsv( v0, v1, v2 );
            else           col->colour().set_rgb( v0, v1, v2 );

            delete d0;
            delete d1;
            delete d2;

            return col;
        }// else
    }
};

//
// ambient/diffuse/specular coefficients
//
class TColCoeffNode : public TParser::TNode
{
protected:
    TParser::TNode * _coeff;
    TParser::TNode * _degree;
    uint             _type;

public:
    TColCoeffNode ( TParser::TNode * coeff,
                    TParser::TNode * deg,
                    uint             type )
            : _coeff(coeff), _degree(deg), _type(type)
    {}
    
    virtual uint type () const
    {
        if (( _coeff != NULL ) && _coeff->is_type( TYPE_NUMBER ))
        {
            if (( _type == TEX_AMBIENT )   || ( _type == TEX_DIFFUSE ) ||
                ( _type == TEX_TRANSMISS ) || ( _type == TEX_IOR ))
                return TYPE_COL_COEFF;

            if (( _type == TEX_SPECULAR ) && ( _degree != NULL ) &&
                _degree->is_type( TYPE_NUMBER ))
                return TYPE_COL_COEFF;

            return TYPE_UNDEF;
        }// if
        else
            return TYPE_UNDEF;
    }

    TParser::TNodeData * eval () const
    {
        if ( _type >= MAX_TEX_VAL )
        {
            parse_error( "unkown coefficient type", pos() );
            return NULL;
        }// if

        if ( _coeff == NULL )
        {
            parse_error( "have NULL coefficient", pos() );
            return NULL;
        }// if

        if (( _type == TEX_SPECULAR ) && ( _degree == NULL ))
        {
            parse_error( "have NULL specular degree", pos() );
            return NULL;
        }// if

        TParser::TNodeData * cd = _coeff->eval();

        if ( cd == NULL )
        {
            parse_error( "got NULL in eval", _coeff->pos() );
            return NULL;
        }// if

        if ( ! cd->is_type( TYPE_NUMBER ) )
        {
            parse_error( "expected number as first argument", _coeff->pos() );
            delete cd;
            return NULL;
        }// if

        real  coefficient = ((TNDNumber*) cd)->number();

        delete cd;

        if ( _type == TEX_SPECULAR )
        {
            TParser::TNodeData * dd = _degree->eval();

            if ( dd == NULL )
            {
                parse_error( "got NULL in eval", _degree->pos() );
                return NULL;
            }// if
            
            if ( ! dd->is_type( TYPE_NUMBER ) )
            {
                parse_error( "expected number as second argument", _degree->pos() );
                delete dd;
                return NULL;
            }// if

            real  degree = ((TNDNumber*) dd)->number();

            delete dd;
            
            return new TNDColCoeff( _type, coefficient, degree );
        }// if
        else
        {
            return new TNDColCoeff( _type, coefficient );
        }// else

        return NULL;
    }
};

//
// composite texture (list of the above)
//
class TTextureNode : public TParser::TNode
{
protected:
    // list of single texture definitions
    TSLL< TParser::TNode * >  _tex_list;

public:
    TTextureNode () {}

    ~TTextureNode ()
    {
        while ( _tex_list.size() > 0 )
            delete _tex_list.behead();
    }
    
    virtual uint type () const { return TYPE_TEXTURE; }

    void append ( TParser::TNode * n )
    {
        if ( n != NULL )
            _tex_list.append( n );
        else
            parse_error( "appending NULL node to texture-list", pos() );
    }
    
    TParser::TNodeData * eval () const
    {
        TNDTexture  * tex = new TNDTexture;

        TSLL< TParser::TNode * >::TIterator  iter = _tex_list.first();

        while ( ! iter.eol() )
        {
            TParser::TNodeData * dat = iter()->eval();

            if ( dat->is_type( TYPE_COLOUR ) )
            {
                TNDColour      * col;

                if (( col = dynamic_cast< TNDColour * >( dat ) ) != NULL )
                {
                    if ( tex->has_tex_val( TEX_COLOUR ) )
                        parse_warn( "colour already defined in texture", iter()->pos() ) ;
                    
                    tex->texture()->set_colour( col->colour() );
                    tex->set_tex_val( TEX_COLOUR, true );
                }// if
                else
                    parse_error( "unknown texture-type in texture-list",
                                 iter()->pos() );
            }// if
            else if ( dat->is_type( TYPE_COL_COEFF ) )
            {
                TNDColCoeff * coeff = dynamic_cast< TNDColCoeff * >( dat );

                if ( tex->has_tex_val( coeff->tex_type() ) )
                {
                    TString  str_type;
                    
                    switch ( coeff->tex_type() )
                    {
                    case TEX_AMBIENT    : str_type = "ambient"; break;
                    case TEX_DIFFUSE    : str_type = "diffuse"; break;
                    case TEX_SPECULAR   : str_type = "specular"; break;
                    case TEX_TRANSMISS  : str_type = "transmiss"; break;
                    case TEX_IOR        : str_type = "ior"; break;
                    default             : str_type = "unkown"; break;
                    }// switch

                    parse_warn( str_type + " already defined in texture", iter()->pos() );
                }// if
                
                switch ( coeff->tex_type() )
                {
                case TEX_AMBIENT : 
                    tex->texture()->set_ambient( coeff->coeff1() ); break;

                case TEX_DIFFUSE :
                    tex->texture()->set_diffuse( coeff->coeff1() ); break;
                    
                case TEX_SPECULAR :
                    tex->texture()->set_specular( coeff->coeff1(), coeff->coeff2() );
                    break;

                case TEX_TRANSMISS :
                    tex->texture()->set_transmiss( coeff->coeff1() ); break;
                    
                case TEX_IOR :
                    tex->texture()->set_ior( coeff->coeff1() ); break;

                default :
                    parse_error( "unkown coefficient-type", iter()->pos() );
                }// switch

                tex->set_tex_val( coeff->tex_type(), true );
            }// if
            else if ( dat->is_type( TYPE_TEXTURE ) )
            {
                TNDTexture  * ndt = (TNDTexture*) dat;

                for ( uint i = 0; i < MAX_TEX_VAL; i++ )
                {
                    if ( ndt->has_tex_val( i ) )
                    {
                        if ( tex->has_tex_val( i ) )
                        {
                            TString  str_type;
                            
                            switch ( i )
                            {
                            case TEX_COLOUR     : str_type = "colour"; break;
                            case TEX_AMBIENT    : str_type = "ambient"; break;
                            case TEX_DIFFUSE    : str_type = "diffuse"; break;
                            case TEX_SPECULAR   : str_type = "specular"; break;
                            case TEX_TRANSMISS  : str_type = "transmiss"; break;
                            case TEX_IOR        : str_type = "ior"; break;
                            default             : str_type = "unkown"; break;
                            }// switch
                            
                            parse_warn( str_type + " already defined in texture", iter()->pos() );
                        }// if

                        switch ( i )
                        {
                        case TEX_COLOUR :
                            tex->texture()->set_colour( ndt->texture()->colour() );
                            break;

                        case TEX_AMBIENT :
                            tex->texture()->set_ambient( ndt->texture()->ambient() );
                            break;
                                
                        case TEX_DIFFUSE :
                            tex->texture()->set_diffuse( ndt->texture()->diffuse() );
                            break;
                                
                        case TEX_SPECULAR :
                            tex->texture()->set_specular( ndt->texture()->specular(),
                                                         ndt->texture()->specular_degree() );
                            break;
                                
                        case TEX_TRANSMISS :
                            tex->texture()->set_transmiss( ndt->texture()->transmiss() );
                            break;
                                
                        case TEX_IOR :
                            tex->texture()->set_ior( ndt->texture()->ior() );
                            break;

                        default :
                            parse_error( "unkown texture-type", iter()->pos() );
                        }// switch
                        
                        tex->set_tex_val( i, true );
                    }// if
                }// for
            }// if
            else
            {
                parse_error( "unknown type in texture-list", iter()->pos() );
                delete tex;
                delete dat;
                return NULL;
            }// else

            delete dat;
            
            ++iter;
        }// while

        return tex;
    }
};

///////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////
//
// return local node types to outer world
//
///////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////

TParser::TNode * new_rgb_colour ( TParser::TNode * n0,
                                  TParser::TNode * n1,
                                  TParser::TNode * n2 )
{
    return new TColourNode( n0, n1, n2, false );
}

TParser::TNode * new_hsv_colour ( TParser::TNode * n0,
                                  TParser::TNode * n1,
                                  TParser::TNode * n2 )
{
    return new TColourNode( n0, n1, n2, true );
}

TParser::TNode * new_colour ( TParser::TNode * n )
{
    return new TColourNode( n );
}

TParser::TNode * new_ambient ( TParser::TNode * n )
{
    return new TColCoeffNode( n, NULL, TEX_AMBIENT );
}

TParser::TNode * new_diffuse ( TParser::TNode * n )
{
    return new TColCoeffNode( n, NULL, TEX_DIFFUSE );
}

TParser::TNode * new_specular ( TParser::TNode * n0, TParser::TNode * n1 )
{
    return new TColCoeffNode( n0, n1, TEX_SPECULAR );
}

TParser::TNode * new_transmiss ( TParser::TNode * n )
{
    return new TColCoeffNode( n, NULL, TEX_TRANSMISS );
}

TParser::TNode * new_ior ( TParser::TNode * n )
{
    return new TColCoeffNode( n, NULL, TEX_IOR );
}

TParser::TNode * new_texture ()
{
    return new TTextureNode();
}

void add_texture ( TParser::TNode * n0, TParser::TNode * n1 )
{
    TTextureNode * tex = dynamic_cast< TTextureNode * >( n0 );

    if ( tex == NULL )
    {
        if ( n0 != NULL )
            parse_error( "no texture-node to add to", n0->pos() );
        else
            parse_error( "no texture-node to add to" );
    }// if

    tex->append( n1 );
}

void apply_colour ( TGraphObject * obj, TParser::TNodeData * nd )
{
    TNDColour     * col;
    TNDColCoeff  * coeff;
    TNDTexture   * tex;

    if (( col = dynamic_cast< TNDColour * >( nd ) ) != NULL )
        obj->texture()->set_colour( col->colour() );
    else if (( coeff = dynamic_cast< TNDColCoeff * >( nd ) ) != NULL )
    {
        switch ( coeff->tex_type() )
        {
        case TEX_AMBIENT :
            obj->texture()->set_ambient( coeff->coeff1() );
            break;
                
        case TEX_DIFFUSE :
            obj->texture()->set_diffuse( coeff->coeff1() );
            break;
                
        case TEX_SPECULAR :
            obj->texture()->set_specular( coeff->coeff1(), coeff->coeff2() );
            break;
                
        case TEX_TRANSMISS :
            obj->texture()->set_transmiss( coeff->coeff1() );
            break;
                
        case TEX_IOR :
            obj->texture()->set_ior( coeff->coeff1() );
            break;

        default:
            parse_error( "unkown texture-type" );
        }// switch
    }// if
    else if (( tex = dynamic_cast< TNDTexture * >( nd ) ) != NULL )
    {
        obj->set_texture( tex->texture() );
        
//         if ( tex->has_tex_val( TEX_COLOUR ) )
//             obj->texture()->set_colour( tex->texture()->colour() );
//         if ( tex->has_tex_val( TEX_AMBIENT ) )
//             obj->texture()->set_ambient( tex->texture()->ambient() );
//         if ( tex->has_tex_val( TEX_DIFFUSE ) )
//             obj->texture()->set_diffuse( tex->texture()->diffuse() );
//         if ( tex->has_tex_val( TEX_SPECULAR ) )
//             obj->texture()->set_specular( tex->texture()->specular(),
//                                           tex->texture()->specular_degree() );
//         if ( tex->has_tex_val( TEX_TRANSMISS ) )
//             obj->texture()->set_transmiss( tex->texture()->transmiss() );
//         if ( tex->has_tex_val( TEX_IOR ) )
//             obj->texture()->set_ior( tex->texture()->ior() );
    }// if
    else
        parse_error( "unknown texture-type to apply" );
}

TColour & to_colour ( TParser::TNodeData * nd )
{
    if ( ! nd->is_type( TYPE_COLOUR ) )
        parse_error( "no colour-data found" );
    
    return ((TNDColour *) nd)->colour();
}

}// namespace
