//
// Project : lib3D
// File    : TImage.cc
// Purpose : class for hardware-independent images
// Author  : Ronald Kriemann
//
// arch-tag: 7a1f2aa6-8c91-446b-8c09-9fbdacd2d48f
//

#include <string.h>
#include <limits.h>

#include <iostream>
#include <cmath>

#include "TImage.hh"

namespace RGL
{

//////////////////////////////////////////////////////
//
// access local variables
//

void
TImage::set_geometry ( uint w, uint h )
{
    if (( w != _width ) || ( h != _height))
    {
        _width  = w;
        _height = h;
        _image.set_size( _width * _height );
    }// if
}

//////////////////////////////////////////////////
//
// various drawing-primitives
//

//
// fill image with colour
//
void
TImage::fill ( const TColour & c )
{
    for ( uint j = 0; j < _height; j++ )
        for ( uint i = 0; i < _width; i++ )
            set_colour( i, j, c );
}

//
// copy given image at given position
//
void
TImage::bitblt ( uint x, uint y, const TImage & img )
{
    uint  w = img.width();
    uint  h = img.height();
    
    for ( uint j = 0; j < h; j++ )
        for ( uint i = 0; i < w; i++ )
            set_colour( i+x, j+y, img.get_colour( i, j ) );
}

//////////////////////////////////////////////////////
//
// image-manipulation
//

//
// convert image to grayscale
//
void
TImage::to_grayscale ()
{
    //
    // go over image and convert RGB-values into grayscale-value
    //
    
    for ( uint j = 0, k = 0; j < _height; j++ )
    {
        for ( uint i = 0; i < _width; i++, k++ )
        {
            TColour  col( get_colour( i, j ) );
            real     gray;

#if 0
            // older transformation
            gray  = col.red()   * 0.30;
            gray += col.green() * 0.59;
            gray += col.blue()  * 0.11;
#else
            // new one for CRT and HDTV phosphors
            // (see Akenine-Mller and Haines, "Real-Time Rendering", chap. 6.2
            gray  = col.red()   * 0.2125;
            gray += col.green() * 0.7154;
            gray += col.blue()  * 0.0721;
#endif
            col.set_rgb( gray, gray, gray );
            set_colour( i, j, col );
        }// for
    }// for
}

//
// make image lighter / darker
//
void
TImage::darker ()
{
    TColour col;

    for ( uint j = 0; j < _height; j++ )
    {
        for ( uint i = 0; i < _width; i++ )
        {
            col = get_colour( i, j );
            col.scale( 0.9 );
            
            set_colour( i, j, col );
        }// for
    }// for
}

void
TImage::lighter ()
{
    TColour col;

    for ( uint j = 0; j < _height; j++ )
    {
        for ( uint i = 0; i < _width; i++ )
        {
            col = get_colour( i, j );
            col.scale( 10.0 / 9.0 );
            
            set_colour( i, j, col );
        }// for
    }// for
}
    
//////////////////////////////////////////////////////
//
// misc.
//

//
// copy methods
//
TImage &
TImage::copy ( const TImage & img )
{
    _width  = img._width;
    _height = img._height;
    
    _image.set_size( _width * _height );

    for ( uint i = 0; i < _width; i++ )
    {
        for ( uint j = 0; j < _height; j++ )
            set_colour( i, j, img.get_colour( i, j ) );
    }// for

    return *this;
}

} // namespace

