//
// Project : lib3D
// File    : TVRML.cc
// Purpose : class for a vrml-printer
// Author  : Ronald Kriemann
//
// arch-tag: 344b7ae7-6291-470d-b00a-fce09505d330
//

#include "TVRML.hh"

namespace RGL
{

#define SUPER  TPrinter

//////////////////////////////////////////////////
//
// constructor and destructor
//

TVRML::TVRML ()
{
    _file     = NULL;
    _filename = "";
    _version  = VRML1;
    _active   = false;
}

TVRML::~TVRML ()
{
    if ( _active )
        end();
}
    
//////////////////////////////////////////////////
//
// methods for drawing
//

// begin and end drawing
void
TVRML::begin ( const char * name )
{
    _filename = name;
    begin();
}

void
TVRML::begin ()
{
    if ( _active )
    {
        std::cout << "TVRML::begin : still active" << std::endl;
        return;
    }// if
    
    if (( _file = fopen( _filename.c_str(), "w" )) == NULL)
    {
        std::cout << "TVRML::begin : could not open file ( " << _filename << ")"
                  << std::endl;
        return;
    }// if

    _active = true;
    _ofs    = 0;

    //
    // write header of vrml file
    //

    if ( _version == VRML1 )
    {
        write( "#VRML V1.0 ascii\n" );
        begin_section();
        
        // set initial camera
        //        beginNode( "OrthographicCamera" );
        //        indentWrite( "position       0 0 1\n" );
        //        indentWrite( "orientation    0 0 1  0\n" );
        //        indentWrite( "focalDistance  5\n" );
        //        indentWrite( "height         2\n" );
        //        endNode();
    }// if
    else
        std::cout << "TVRML::begin : not supported vrml-version" << std::endl;
}

void
TVRML::end ()
{
    if ( ! _active )
        return;

    end_section();

    fclose( _file );
    _file   = NULL;
    _active = false;
}

//
// 3D - drawing
//

void
TVRML::draw_point ( const TVector & v )
{
}

void
TVRML::draw_line ( const TVector & orig, const TVector & dest )
{
    char     buffer[256];
    TVector  d;

    // avoid zero-length lines
    d.set( 1.0, orig, -1.0, dest );
    if ( d.norm2() < EPSILON )
        return;
    
    begin_section();

    write_colour();
    
    begin_section( "Coordinate3" );
    write( "point [" );
    
    snprintf( buffer, sizeof(buffer), "%.4g %.4g %.4g,",
              double(orig.x()), double(orig.y()), double(-orig.z()) );
    write( buffer );

    snprintf( buffer, sizeof(buffer), "%.4g %.4g %.4g,",
              double(dest.x()), double(dest.y()), double(-dest.z()) );
    write( buffer );

    write( "]" );
    end_section();
    
    begin_section( "IndexedLineSet" );
    write( "coordIndex [ 0, 1, -1 ]" );
    end_section();

    end_section();
}

//
// special methods for drawing polygons
//
void
TVRML::draw_triangle ( const TVector & x1, const TVector & x2,
                       const TVector & x3 )
{
    draw_line( x1, x2 );
    draw_line( x2, x3 );
    draw_line( x3, x1 );
}

void
TVRML::fill_triangle ( const TVector & x1, const TVector & x2,
                       const TVector & x3 )
{
}

void
TVRML::draw_rect ( const TVector & x1, const TVector & x2,
                   const TVector & x3, const TVector & x4 )
{
    draw_line( x1, x2 );
    draw_line( x2, x3 );
    draw_line( x3, x4 );
    draw_line( x4, x1 );
}

void
TVRML::fill_rect ( const TVector & x1, const TVector & x2,
                   const TVector & x3, const TVector & x4 )
{
}

//
// draw polygon by giving an array of points
//
void
TVRML::draw_poly ( TVector * points, int count )
{
}

void
TVRML::fill_poly ( TVector * points, int count )
{
}

//
// draw a sphere
//
void
TVRML::draw_sphere ( const TVector & center, real radius )
{
    char  buffer[256];

    begin_section();

    write_colour();
    
    begin_section( "Translation" );
    snprintf( buffer, sizeof(buffer),
              "translation %.2g %.2g %.2g",
              double(  center.x() ),
              double(  center.y() ),
              double( -center.z() ) );
    write( buffer );
    end_section();

    begin_section( "Sphere" );
    snprintf( buffer, sizeof(buffer), "radius %.4g", double( radius ) );
    write( buffer );
    end_section();

    end_section();
}
    
//////////////////////////////////////////////////
//
// methods for setting up style
//

//
// set up texture for all following objects
//
void
TVRML::set_texture ( const TTexture & t )
{
    _texture = t;
}

//////////////////////////////////////////////////
//
// misc helping-methods
//

void
TVRML::begin_section ( const char * name )
{
    char  buffer[256];

    if (name != NULL)
        snprintf( buffer, sizeof(buffer), "%s {", name );
    else
        strncpy( buffer, "Separator {", sizeof(buffer) );
    
    write( buffer );
    _ofs += 2;
}

void
TVRML::end_section ()
{
    write( "}" );
    _ofs -= 2;
}

void
TVRML::write ( const char * text )
{
    for ( uint i = 0; i < _ofs; i++ )
        fputc( ' ', _file );
    
    fputs( text, _file );
    fputc( '\n', _file );
}

void
TVRML::write_colour ()
{
    char  buffer[256];
    
    begin_section( "Material" );

    snprintf( buffer, sizeof(buffer),
              "diffuseColor %.2g %.2g %.2g",
              double( texture().colour().red() ),
              double( texture().colour().green() ),
              double( texture().colour().blue() ) );
    write( buffer );
    
//     if ((colour().amb_r != 0) || (colour().amb_g != 0) || (colour().amb_b != 0))
//     {
//         sprintf( buffer, "ambientColor %.4g %.4g %.4g",
//                  double( texture().colour().amb_r ),
//                  double( texture().colour().amb_g ),
//                  double( texture().colour().amb_b ) );
//         indentWrite( buffer );
//     }// if
    
    end_section();
}

} // namespace

