//
// Project : lib3D
// File    : TAdaSuperSampRT.cc
// Purpose : class for raytracing by super sampling
// Author  : Ronald Kriemann
//
// arch-tag: 438cd5c6-bc51-4f5f-8f13-d75c9d7ca16f
//

#include "TRNG.hh"
#include "TAdaSuperSampRT.hh"

namespace RGL
{

///////////////////////////////////////////////
//
// constructor and destructor
//

TAdaSuperSampRT::TAdaSuperSampRT ()
{
    _min_contrast = 0.01;
}

///////////////////////////////////////////////
//
// rendering
//

void
TAdaSuperSampRT::render_tile ( const TTileDesc & tile,
                               TImage          & img,
                               TProgressBar    * bar ) const
{
    //
    // we need three pixel lines to access all neighbours
    // of a pixel
    //
    
    uint                 width  = tile.width;
    uint                 height = tile.height;
    TArray< TColour * >  pix_line1( width );  // last line
    TArray< TColour * >  pix_line2( width );  // current line
    TArray< TColour * >  pix_line3( width );  // next line

    for ( uint i = 0; i < width; i++ )
    {
        pix_line1[i] = new TColour;
        pix_line2[i] = new TColour;
        pix_line3[i] = new TColour;
    }// for
    
    //
    // go over tile and compute colour for each pixel
    //

    TArray< TRay >   rays;
    TColour          colour, pix_colour;
    TArray< real >   x_distr;
    TArray< real >   y_distr;
    
    if ( verbose(1) && ( bar != NULL ))
        bar->init( 0, height, 0 );
    
    for ( uint j = 0; j < height; j++ )
    {
        for ( uint i = 0; i < width; i++ )
        {
            //
            // compute colour of pixel
            //

            camera()->set_rays( rays,
                                (real(i+tile.x_ofs) + 0.5) / real(tile.iwidth),
                                (real(j+tile.y_ofs) + 0.5) / real(tile.iheight) );

            for ( uint l = 0; l < rays.size(); l++ )
            {
                TColour  c;
                
                trace_ray( rays[l], 0, 1.0, c );

                pix_line2[i]->add( c );
            }// for
            
            pix_line2[i]->scale( 1.0 / real(rays.size()) );
        }// for

        //
        // compare pixel with neighbours and supersample
        // if contrast is too high
        //

        if ( j > 1 )
        {
            TColour * upper, * lower, * left, * right;
            
            for ( uint i = 0; i < width; i++ )
            {
                if ( j > 2 ) upper = pix_line1[i];
                else         upper = NULL;

                if ( j < height-1 ) lower = pix_line3[i];
                else                lower = NULL;

                if ( i > 0 ) left = pix_line2[i-1];
                else         left = NULL;

                if ( i < width-1 ) right = pix_line2[i+1];
                else               right = NULL;

                if ( exceeds_contrast( pix_line2[i], upper, lower, left, right ) )
                {
                    //
                    // supersample pixel
                    //

                    pix_colour.set( * pix_line2[i] );
                    
                    build_distr( x_distr, y_distr );
                    
                    for ( uint l = 0; l < x_distr.size(); l++ )
                    {
                        //
                        // get ray from camera, compute colour of sub-pixel
                        // and add to pixel-colour
                        //
                        
                        camera()->set_rays( rays,
                                            (real(i+tile.x_ofs) + x_distr[l])/real(tile.iwidth),
                                            (real(j+tile.y_ofs) + y_distr[l])/real(tile.iheight) );
                        
                        trace_ray( rays[l], 0, 1.0, colour );
                        
                        pix_colour.add( colour, false );
                    }// for

                    pix_colour.scale( 1.0 / real(_no_of_samples+1) );

                    pix_line2[i]->set( pix_colour );
                }// if

                //
                // compute and set colour of pixel
                //
                
                img.draw_point( i, j, * pix_line2[i] );
            }// for
        }// if

        //
        // shift pixel-lines
        //
        
        for ( uint i = 0; i < width; i++ )
        {
            TColour * tmp = pix_line1[i];

            pix_line1[i] = pix_line2[i];
            pix_line2[i] = pix_line3[i];
            pix_line3[i] = tmp;
        }// for

        if ( verbose(1) && ( bar != NULL ))
            bar->advance( 1 );
    }// for

    //
    // clean up
    //
    
    for ( uint i = 0; i < width; i++ )
    {
        delete pix_line1[i];
        delete pix_line2[i];
        delete pix_line3[i];
    }// for
    
    if ( verbose(1) && ( bar != NULL ))
        bar->finish();
}

///////////////////////////////////////////////
//
// determine contrast 
//

bool
TAdaSuperSampRT::exceeds_contrast ( const TColour * pix,
                                    const TColour * upper,
                                    const TColour * lower,
                                    const TColour * left,
                                    const TColour * right ) const
{
    assert( pix != NULL );

    //
    // compare colour with neighbours
    //

    if (( upper != NULL ) && ( colour_dist( *pix, *upper ) > _min_contrast ))
        return true;

    if (( lower != NULL ) && ( colour_dist( *pix, *lower ) > _min_contrast ))
        return true;

    if (( left != NULL ) && ( colour_dist( *pix, *left ) > _min_contrast ))
        return true;

    if (( right != NULL ) && ( colour_dist( *pix, *right ) > _min_contrast ))
        return true;

    return false;
}

} // namespace
