//
// Project : lib3D
// File    : TAdaptiveRT.cc
// Purpose : class for raytracing by super sampling
// Author  : Ronald Kriemann
//
// arch-tag: e34bd831-9495-4800-977a-63e0f584ea87
//

#include "TRNG.hh"
#include "TArray.hh"
#include "TMultiRef.hh"
#include "TAdaptiveRT.hh"

namespace RGL
{

///////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////
//
// represents a single pixel which can be subdivied
//
// use to compute the final colour of a pixel by
// summing up colours of subpixel
//
///////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////

enum { PIX_UL = 1,
       PIX_UR = 2,
       PIX_LL = 4,
       PIX_LR = 8 };

class TPixel : public TColour
{
protected:
    // upper-left corner and stepwidth of pixel
    TVector   _ulc;
    TVector   _step_x, _step_y;
    
    // subpixel
    TPixel  * _sub_pixel[4];

    // true if colour was computed
    bool      _computed;
    
public:
    //
    // constructor
    //
    
    TPixel ()
    {
        _sub_pixel[0] = _sub_pixel[1] = _sub_pixel[2] = _sub_pixel[3] = NULL;
    }

    TPixel ( const TVector & ulc,
             const TVector & step_x,
             const TVector & step_y )
            : _ulc(ulc), _step_x(step_x), _step_y(step_y)
    {
        _sub_pixel[0] = _sub_pixel[1] = _sub_pixel[2] = _sub_pixel[3] = NULL;
    }

    virtual ~TPixel ()
    {
        coarsen();
    }
    
    //
    // coordinates of pixel
    //

    TVector & ulc    () { return _ulc; }
    TVector & step_x () { return _step_x; }
    TVector & step_y () { return _step_y; }

    //
    // access sub-pixels
    //

    TPixel * sub_pixel ( uint i )
    {
        switch ( i )
        {
        case PIX_UL : return _sub_pixel[0];
        case PIX_UR : return _sub_pixel[1];
        case PIX_LL : return _sub_pixel[2];
        case PIX_LR : return _sub_pixel[3];
        }// switch

        std::cout << "(TPixel) sub_pixel : wrong index for sub-pixel" << std::endl;
        return NULL;
    }
    
    void set_sub_pixel ( uint i, TPixel * p )
    {
        switch ( i )
        {
        case PIX_UL : _sub_pixel[0] = p; break;
        case PIX_UR : _sub_pixel[1] = p; break;
        case PIX_LL : _sub_pixel[2] = p; break;
        case PIX_LR : _sub_pixel[3] = p; break;
        default     :
            std::cout << "(TPixel) set_sub_pixel : wrong index for sub-pixel" << std::endl;
        }// switch
    }
    
    // create sub-pixels at specified positions
    void refine ( uint pix_mask )
    {
        TVector  sx2, sy2, su;

        sx2.set( 0.5, _step_x );
        sy2.set( 0.5, _step_y );
        
        if (( pix_mask & PIX_UL ) && ( sub_pixel( PIX_UL ) == NULL ))
            set_sub_pixel( PIX_UL, new TPixel( _ulc, sx2, sy2 ) );

        if (( pix_mask & PIX_UR ) && ( sub_pixel( PIX_UR ) == NULL ))
        {
            su.set( 1.0, _ulc, 1.0, sx2 );
            set_sub_pixel( PIX_UR, new TPixel( _ulc, sx2, sy2 ) );
        }// if

        if (( pix_mask & PIX_LL ) && ( sub_pixel( PIX_LL ) == NULL ))
        {
            su.set( 1.0, _ulc, 1.0, sy2 );
            set_sub_pixel( PIX_LL, new TPixel( _ulc, sx2, sy2 ) );
        }// if
        
        if (( pix_mask & PIX_LR ) && ( sub_pixel( PIX_LR ) == NULL ))
        {
            su.set( 1.0, _ulc).add( 1.0, sx2, 1.0, sy2 );
            set_sub_pixel( PIX_LR, new TPixel( _ulc, sx2, sy2 ) );
        }// if
    }

    // remove all sub-pixels
    void coarsen ()
    {
        delete _sub_pixel[0];
        delete _sub_pixel[1];
        delete _sub_pixel[2];
        delete _sub_pixel[3];

        _sub_pixel[0] = _sub_pixel[1] = _sub_pixel[2] = _sub_pixel[3] = NULL;
    }
    
    //
    // access computation flag
    //
    
    bool is_computed () const { return _computed; }

    void set_computed ( bool b ) { _computed = b; }
};

///////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////
//
// TAdaptiveRT implementation
//
///////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////

///////////////////////////////////////////////
//
// constructor
//

TAdaptiveRT::TAdaptiveRT ()
{
    _max_contrast = 0.01;
    _max_ada_lvl  = 2;
}

///////////////////////////////////////////////
//
// rendering
//

void
TAdaptiveRT::render_tile ( const TTileDesc & tile,
                           TImage          & img,
                           TProgressBar    * bar ) const
{
    uint   width  = img.width();
    uint   height = img.height();

    //
    // start point (upper left corner of the image)
    //

    TVector  ulc, step_x, step_y;

    step_x.set( 2.0 / real(tile.iwidth), camera()->right() );
    step_y.set( -2.0 / real(tile.iheight), camera()->up() );

    ulc.set( 1.0, camera()->look_at() );
    ulc.add( 1.0, camera()->up(), -1.0, camera()->right() );
    ulc.add( tile.y_ofs, step_y, tile.x_ofs, step_x );

    //
    // we need three pixel lines to access all neighbours
    // of a pixel
    //
    
    TArray< TPixel * >  pix_line1( width );  // last line
    TArray< TPixel * >  pix_line2( width );  // current line
    TArray< TPixel * >  pix_line3( width );  // next line
    TVector             dir;

    for ( uint i = 0; i < width; i++ )
    {
        pix_line1[i] = new TPixel;
        pix_line2[i] = new TPixel;
        pix_line3[i] = new TPixel;
    }// for
    
    //
    // go over image and shoot ray in the middle of each pixel
    // if colour differs too much with neighbours: subdivide
    //

    TPixel * upper, * lower, * left, *right;
    
    if ( verbose(1) && ( bar != NULL ))
        bar->init( 0, height, 0 );

    for ( uint j = 0; j < height; j++ )
    {
        //
        // compute beginning of image-line in world-space
        //

        dir.set( 1.0, ulc, -1.0, camera()->origin() );
            
        for ( uint i = 0; i < width; i++ )
        {
            //
            // compute colour for given pixel
            //

            pix_line3[i]->ulc()    = dir;
            pix_line3[i]->step_x() = step_x;
            pix_line3[i]->step_y() = step_y;
        
            render_pixel( pix_line3[i] );

            //
            // set color of pixel
            //

            img.draw_point( i, j, * pix_line3[i] );

            //
            // update lookat
            //
            
            dir.add( 1.0, step_x );
        }// for

        //
        // compare neighboured pixel and refine if neccessary
        //

        if ( j > 2 )
        {
            for ( uint i = 0; i < width; i++ )
            {
                if ( j > 2 ) upper = pix_line1[i];
                else         upper = NULL;

                if ( j < height-1 ) lower = pix_line3[i];
                else                lower = NULL;

                if ( i > 0 ) left = pix_line2[i-1];
                else         left = NULL;

                if ( i < width-1 ) right = pix_line2[i+1];
                else               right = NULL;

                compare_pixel( pix_line2[i], upper, lower, left, right, 0 );
            }// for
        }// if

        //
        // shift pixel-lines
        //
        
        for ( uint i = 0; i < width; i++ )
        {
            TPixel * tmp = pix_line1[i];

            pix_line1[i] = pix_line2[i];
            pix_line2[i] = pix_line3[i];
            pix_line3[i] = tmp;

            tmp->coarsen();
            tmp->set_computed( false );
        }// for
        
        //
        // update upper-left corner
        //

        ulc.add( 1.0, step_y );

        if ( verbose(1) && ( bar != NULL ))
            bar->advance( 1 );
    }// for

    //
    // clean up
    //
    
    for ( uint i = 0; i < width; i++ )
    {
        delete pix_line1[i];
        delete pix_line2[i];
        delete pix_line3[i];
    }// for
    
    if ( verbose(1) && ( bar != NULL ))
        bar->finish();
}

///////////////////////////////////////////////
//
// adaptivity methods
//

//
// adaptivly render colour for a pixel
//
void
TAdaptiveRT::render_pixel ( TPixel * pixel ) const
{
    //
    // initial tests
    //
    
    if ( pixel == NULL )
        std::cout << "(TAdaptiveRT) render_pixel : supplied pixel is NULL" << std::endl;

    if ( pixel->is_computed() )
        return;
    
    //
    // shoot ray through the center of the pixel
    //

    TRay  ray( true );

    ray.origin().set( 1.0, camera()->origin() );
    ray.direction().set( 1.0, pixel->ulc() );
    ray.direction().add( 0.5, pixel->step_x(), 0.5, pixel->step_y() );
    
    pixel->set_rgb( 0, 0, 0 );
    trace_ray( ray, 0, 1.0, * pixel );

    pixel->set_computed( true );
}

//
// compare pixel with neighbours
//
void
TAdaptiveRT::compare_pixel ( TPixel * pixel,
                             TPixel * upper,
                             TPixel * lower,
                             TPixel * left,
                             TPixel * right,
                             uint     lvl ) const
{
    assert( pixel != NULL );

    if ( lvl >= _max_ada_lvl )
        return;
    
    //
    // compare colour with neighboured colours and
    // refine if neccessary
    //

    bool refined = false;
    
    if (( upper != NULL ) && ( colour_dist( * pixel, * upper ) > _max_contrast ))
    {
        refined = true;
        
        pixel->refine( PIX_UL | PIX_UR );
        upper->refine( PIX_LL | PIX_LR );

        render_pixel( pixel->sub_pixel( PIX_UL ) );
        render_pixel( upper->sub_pixel( PIX_LL ) );
        compare_pixel( pixel->sub_pixel( PIX_UL ), NULL, NULL, NULL,
                       upper->sub_pixel( PIX_LL ), lvl+1 );
        
        render_pixel( pixel->sub_pixel( PIX_UR ) );
        render_pixel( upper->sub_pixel( PIX_LR ) );
        compare_pixel( pixel->sub_pixel( PIX_UR ), NULL, NULL, NULL,
                       upper->sub_pixel( PIX_LR ), lvl+1 );
    }// if
    
    if (( lower != NULL ) && ( colour_dist( * pixel, * lower ) > _max_contrast ))
    {
        refined = true;
        
        pixel->refine( PIX_LL | PIX_LR );
        lower->refine( PIX_UL | PIX_UR );

        render_pixel( pixel->sub_pixel( PIX_LL ) );
        render_pixel( lower->sub_pixel( PIX_UL ) );
        compare_pixel( pixel->sub_pixel( PIX_LL ), NULL, NULL, NULL,
                       lower->sub_pixel( PIX_UL ), lvl+1 );
        
        render_pixel( pixel->sub_pixel( PIX_LR ) );
        render_pixel( lower->sub_pixel( PIX_UR ) );
        compare_pixel( pixel->sub_pixel( PIX_LR ), NULL, NULL, NULL,
                       lower->sub_pixel( PIX_UR ), lvl+1 );
    }// if
    
    if (( left != NULL ) && ( colour_dist( * pixel, * left ) > _max_contrast ))
    {
        refined = true;
        
        pixel->refine( PIX_UL | PIX_LL );
        left->refine(  PIX_UR | PIX_LR );

        render_pixel( pixel->sub_pixel( PIX_UL ) );
        render_pixel( left->sub_pixel( PIX_UR ) );
        compare_pixel( pixel->sub_pixel( PIX_UL ), NULL, NULL, NULL,
                       left->sub_pixel( PIX_UR ), lvl+1 );
        
        render_pixel( pixel->sub_pixel( PIX_LL ) );
        render_pixel( left->sub_pixel( PIX_LR ) );
        compare_pixel( pixel->sub_pixel( PIX_LL ), NULL, NULL, NULL,
                       left->sub_pixel( PIX_LR ), lvl+1 );
    }// if
    
    if (( right != NULL ) && ( colour_dist( * pixel, * right ) > _max_contrast ))
    {
        refined = true;
        
        pixel->refine( PIX_UR | PIX_LR );
        right->refine( PIX_UL | PIX_LL );

        render_pixel( pixel->sub_pixel( PIX_UR ) );
        render_pixel( right->sub_pixel( PIX_UL ) );
        compare_pixel( pixel->sub_pixel( PIX_UR ), NULL, NULL, NULL,
                       right->sub_pixel( PIX_UL ), lvl+1 );
        
        render_pixel( pixel->sub_pixel( PIX_LR ) );
        render_pixel( right->sub_pixel( PIX_LL ) );
        compare_pixel( pixel->sub_pixel( PIX_LR ), NULL, NULL, NULL,
                       right->sub_pixel( PIX_LL ), lvl+1 );
    }// if

    //
    // adjust colour in pixel by averaging colour in sub-pixel
    //

    if ( refined )
    {
        TColour  avg;

        if ( pixel->sub_pixel( PIX_UL ) != NULL )
            avg.add( * pixel->sub_pixel( PIX_UL ), false );
        else
            avg.add( * pixel, false );

        if ( pixel->sub_pixel( PIX_UR ) != NULL )
            avg.add( * pixel->sub_pixel( PIX_UR ), false );
        else
            avg.add( * pixel, false );

        if ( pixel->sub_pixel( PIX_LL ) != NULL )
            avg.add( * pixel->sub_pixel( PIX_LL ), false );
        else
            avg.add( * pixel, false );

        if ( pixel->sub_pixel( PIX_LR ) != NULL )
            avg.add( * pixel->sub_pixel( PIX_LR ), false );
        else
            avg.add( * pixel, false );

        avg.scale( 0.25 );
        pixel->set( avg );
    }// if
}

} // namespace
