//
// Project : lib3D
// File    : TRaytracer.cc
// Purpose : baseclass for all raytracers
// Author  : Ronald Kriemann
//
// arch-tag: 76947cad-7309-43ee-a488-9ae2c5914473
//

#include <assert.h>

#include "TLight.hh"
#include "TOctTreeIntersect.hh"
#include "TProgressBar.hh"

#include "TRaytracer.hh"

namespace RGL
{

//
// defines
//
#define INC_COUNTER( m )     (m)++
#define SET_COUNTER( m, v )  (m) = (v)

///////////////////////////////////////////////
//
// constructor and destructor
//

TRaytracer::TRaytracer ()
{
    _world      = NULL;
    _camera     = NULL;
    
    _max_depth     = 10;
    _min_intensity = 0.01;
    _with_shadow   = true;
    
    _background = TColour( "black" );
    
    _intersect  = new TOctTreeIntersect;
}

TRaytracer::~TRaytracer ()
{
    delete _intersect;
}

///////////////////////////////////////////////
//
// rendering
//

void
TRaytracer::render_world ( TImage & img, TProgressBar * bar ) const
{
    // reset ray tracing statistics
    reset_stat();

    //
    // by default the whole image is computed
    // at once
    //

    TTileDesc tile = { 0, 0, img.width(), img.height(), img.width(), img.height() };

    render_tile( tile, img, bar );
}

///////////////////////////////////////////////
//
// raytracing methods
//

//
// recursive raytracing
//
void
TRaytracer::trace_ray ( const TRay & ray,
                        uint         depth,
                        real         intensity,
                        TColour    & colour ) const
{
    //
    // reset colour if initial ray
    //

    if ( ray.is_initial() )
        colour.set_rgb( 0, 0, 0 );
    
    //
    // if maximal depth is reached or the intensity of the
    // ray is below some threshold, we shade it black
    //

    if ( reached_stop( intensity, depth ) )
    {
        colour.set_rgb( 0, 0, 0 );
        return;
    }// if

    INC_COUNTER( _stat.no_of_rays );
    SET_COUNTER( _stat.deepest_level, max(_stat.deepest_level, depth) );
    
    //
    // calulate intersection of ray with objects in world
    //

    real           hit_distance = 0;
    TGraphObject * hit_object   = intersect( ray, hit_distance );

    if ( hit_object != NULL )
    {
        if ( verbose(5) )
            hit_object->print();
        
        THPDesc  hp_desc( hit_object, hit_distance );

        hp_desc.hitpoint().set( 1.0, ray.origin(), hit_distance, ray.direction() );

        // calc normal of hit_object at hit_point
        if ( ! hit_object->calc_normal( hp_desc.hitpoint(), hp_desc.normal() ) )
        {
            std::cout << "(TRaytracer) trace_ray : could not compute normal"
                      << std::endl;
            return;
        }// if
        
        // calculate colour of object hit by ray
        shade( hp_desc, ray, depth, intensity, colour );
    }// if
    else
    {
        //
        // no object was intersected: shade with background colour
        //

        colour.set( _background );
    }// else
}

//
// calculate intersection of ray with objects in world
//
TGraphObject *
TRaytracer::intersect ( const TRay & ray, real & hit_distance ) const
{
    if ( ray.is_initial() )
        return _intersect->intersect( ray, hit_distance,
                                      camera()->front(),
                                      camera()->back() );
    else
        return _intersect->intersect( ray, hit_distance,
                                      EPSILON, -1 );
}

//
// shade object at given point
//
void
TRaytracer::shade ( const THPDesc & hp,
                    const TRay    & ray,
                    uint            depth,
                    real            intensity,
                    TColour       & colour ) const
{
    //
    // go over all lights and illuminate
    //

    TLight   * light;
    TColour    light_colour;

    for ( TWorld::TLightIterator iter = world()->lights().first();
          ! iter.eol();
          ++iter )
    {
        light = iter();

        // check if light is visible from object
        if ( _with_shadow && ! light->is_visible( hp.hitpoint(), * _intersect ) )
            continue;

        // illuminate
        light->illuminate( hp, ray.origin(), light_colour );
        colour.add( light_colour );
    }// for

    //
    // compute reflected and refracted ray and follow them
    //

    real    spec = hp.object()->texture()->specular();
    real    tran = hp.object()->texture()->transmiss();
    TVector incident( ray.direction() );

    // normalise incident direction
    incident.scale( -1.0 / incident.norm2() );

    //
    // follow reflected ray
    //

    if ( ! reached_stop( intensity*spec, depth+1 ) )
    {
        TRay    reflected_ray;
        TColour reflected_colour;

        INC_COUNTER( _stat.reflected_rays );
        
        reflected_ray.copy_history( ray );
        comp_reflected( hp, incident, reflected_ray );
        trace_ray( reflected_ray, depth+1, intensity*spec, reflected_colour );
        reflected_colour.scale( intensity*spec ); 
        colour.add( reflected_colour );
    }// if

    //
    // follow refracted ray (if no total internal reflection occurs)
    //

    if ( ! reached_stop( intensity*tran, depth+1 ) )
    {
        TRay    refracted_ray;
        TColour refracted_colour;
        
        const TGraphObject * medium0 = NULL;
        const TGraphObject * medium1 = NULL;

        //
        // first check if ray enters from atmosphere
        //

        refracted_ray.copy_history( ray );
        
        if ( refracted_ray.outside_objects() )
        {
            medium1 = hp.object();
            refracted_ray.enter( medium1 );
        }// if
        else
        {
            //
            // if inside an object, test if we are entering/leaving
            // the current hitobject
            //

            int nr = -1;
            
            if ( refracted_ray.has_entered( hp.object(), nr ) )
            {
                // ray is leaving object
                refracted_ray.exit( nr );

                if ( refracted_ray.outside_objects() )
                    medium0 = hp.object();
                else
                {
                    medium0 = hp.object();
                    medium1 = refracted_ray.current_object();
                }// else
            }// if
            else
            {
                // ray is entering object
                medium0 = refracted_ray.current_object();
                medium1 = hp.object();
                refracted_ray.enter( medium1 );
            }// else
        }// else
        
        if ( comp_refracted( medium0, medium1, hp, incident, refracted_ray ) )
        {
            INC_COUNTER( _stat.refracted_rays );
        
            trace_ray( refracted_ray, depth+1, intensity*tran, refracted_colour );
            refracted_colour.scale( intensity*tran );
        }// if
        else
        {
            //
            // total internal reflection
            //
            
            INC_COUNTER( _stat.reflected_rays );

            comp_reflected( hp, incident, refracted_ray );
            trace_ray( refracted_ray, depth+1, intensity*spec, refracted_colour );
            refracted_colour.scale( intensity*spec );
            //std::cout << "total internal reflection" << std::endl;
        }// else

        colour.add( refracted_colour );
    }// if
}

//
// compute reflected/refracted ray
//
void
TRaytracer::comp_reflected ( const THPDesc & hp,
                             const TVector & incident,
                             TRay          & ray ) const
{
    ray.origin().set( 1.0, hp.hitpoint() );
    ray.direction().set( 2.0 * dot( hp.normal(), incident ), hp.normal(),
                         -1.0, incident );
}

//
// compute and follow refracted (transmitted) ray
//
bool
TRaytracer::comp_refracted ( const TGraphObject * medium0,
                             const TGraphObject * medium1,
                             const THPDesc      & hp,
                             const TVector      & incident,
                             TRay               & ray ) const
{
    real  ior0 = 1.0;
    real  ior1 = 1.0;
    real  eta;
    real  t, n;
    
    if ( medium0 != NULL ) ior0 = medium0->texture()->ior();
    if ( medium1 != NULL ) ior1 = medium1->texture()->ior();

    eta = ior0 / ior1;

    // check if both media have same ior => only transmit
    if ( std::abs( 1.0 - eta ) < EPSILON )
    {
        ray.origin().set( 1.0, hp.hitpoint() );
        ray.direction().set( -1.0, incident );
        return true;
    }// if

    n = dot( incident, hp.normal() );
    t = 1.0 + eta*eta * ( n*n - 1.0 );

    if ( t < 0.0 )
        return false;

    t = eta * n - std::sqrt( t );
    
    ray.origin().set( 1.0, hp.hitpoint() );
    ray.direction().set( t, hp.normal(), -eta, incident );

    return true;
}

///////////////////////////////////////////////
//
// misc. methods
//

//
// reset statistical information
//
void
TRaytracer::reset_stat () const
{
    _stat.no_of_rays     = 0;
    _stat.reflected_rays = 0;
    _stat.refracted_rays = 0;
    _stat.deepest_level  = 0;
}

}// namespace
