//
// Project : lib3D
// File    : TSuperSamplingRT.cc
// Purpose : class for raytracing by super sampling
// Author  : Ronald Kriemann
//
// arch-tag: 7146c2a3-0358-491a-b7df-870f05af7c69
//

#include <time.h>

#include "TRNG.hh"
#include "TSuperSamplingRT.hh"

namespace RGL
{

///////////////////////////////////////////////
//
// constructor and destructor
//

TSuperSamplingRT::TSuperSamplingRT ()
        : _rng( 1, time( NULL ) )
{
    _pattern       = SS_JITTER;
    _no_of_samples = 1;
}

///////////////////////////////////////////////
//
// rendering
//

void
TSuperSamplingRT::render_tile ( const TTileDesc & tile,
                                TImage          & img,
                                TProgressBar    * bar ) const
{
    //
    // go over tile and compute colour for each pixel
    //

    uint             width  = tile.width;
    uint             height = tile.height;
    TArray< TRay >   rays;
    TColour          colour, pix_colour;
    TArray< real >   x_distr;
    TArray< real >   y_distr;
    
    if ( verbose(1) && ( bar != NULL ))
        bar->init( 0, height, 0 );
    
    for ( uint j = 0; j < height; j++ )
    {
        for ( uint i = 0; i < width; i++ )
        {
            //
            // compute colour for pixel
            //

            pix_colour.set_rgb( 0, 0, 0 );

            build_distr( x_distr, y_distr );
            
            for ( uint l = 0; l < x_distr.size(); l++ )
            {
                //
                // get ray from camera, compute colour of sub-pixel
                // and add to pixel-colour
                //

                camera()->set_rays( rays,
                                    (real(i+tile.x_ofs) + x_distr[l])/real(tile.iwidth),
                                    (real(j+tile.y_ofs) + y_distr[l])/real(tile.iheight) );

                colour.set_rgb( 0, 0, 0 );
                
                for ( uint l = 0; l < rays.size(); l++ )
                {
                    TColour  c;
                    
                    trace_ray( rays[l], 0, 1.0, c );
                
                    colour.add( c, false );
                }// for

                colour.scale( 1.0 / real(rays.size()) );
                pix_colour.add( colour, false );
            }// for

            pix_colour.scale( 1.0 / real(_no_of_samples) );
            
            //
            // compute and set colour of pixel
            //

            img.draw_point( i, j, pix_colour );
        }// for

        if ( verbose(1) && ( bar != NULL ))
            bar->advance( 1 );
    }// for

    if ( verbose(1) && ( bar != NULL ))
        bar->finish();
}

//
// build stochastic distribution for anti-aliasing
//
void
TSuperSamplingRT::build_distr ( TArray< real > & x_distr,
                                TArray< real > & y_distr ) const
{
    if ( _pattern == SS_JITTER )
    {
        //
        // special case for one ray / pixel
        //
    
        if ( _no_of_samples < 4 )
        {
            x_distr.set_size( 1 );
            y_distr.set_size( 1 );

            x_distr[0] = 0.5;
            y_distr[0] = 0.5;

            return;
        }// if

        //
        // build jitter distribution
        //

        uint   n;
        
        n = uint( std::sqrt( real(_no_of_samples) ) );

        x_distr.set_size( n*n );
        y_distr.set_size( n*n );

        real  h  = 1.0 / real(n);
        real  h2 = h / 2.0;
        int   c  = 0;

        for ( uint i = 0; i < n; i++ )
        {
            real x = h2 + (i*h);
        
            for ( uint j = 0; j < n; j++ )
            {
                real  y = h2 + (j*h);

                x_distr[ c ] = x + (_rng(h) - h2);
                y_distr[ c ] = y + (_rng(h) - h2);
                c++;
            }// for
        }// for
    }// if

    if ( _pattern == SS_POISSON )
    {
        //
        // special case for one ray / pixel
        //
    
        if ( _no_of_samples <= 1 )
        {
            x_distr.set_size( 1 );
            y_distr.set_size( 1 );

            x_distr[0] = 0.5;
            y_distr[0] = 0.5;

            return;
        }// if

        //
        // build poisson-disc-distribution
        //
    
        real   min_dist = 0.8 / real(_no_of_samples);
        real   mds      = min_dist * min_dist;

        x_distr.set_size( _no_of_samples );
        y_distr.set_size( _no_of_samples );
    
        for ( uint i = 0; i < _no_of_samples; i++ )
        {
            //
            // build new index and check if min-dist is valid
            //
        
            bool valid = false;
            
            while ( ! valid )
            {
                real x = _rng();
                real y = _rng();
            
                valid = true;
            
                for ( uint j = 0; j < i; j++ )
                {
                    real dx = x - x_distr[j];
                    real dy = y - y_distr[j];
                
                    if ( dx*dx + dy*dy < mds )
                    {
                        valid = false;
                        break;
                    }// if
                }// for
            
                if ( valid )
                {
                    x_distr[i] = x;
                    y_distr[i] = y;
                }// if
            }// while
        }// for
    }// if
}

} // namespace
